package fluids;

/**
 * A <code>Pipe</code> carries water between two <code>Tank</code>s, referred to as its
 * <code>inlet</code> and <code>outlet</code>. Its only other property is the flow rate:
 * how fast water is flowing through the pipe. Flow rate is measured in gallons per second.
 * A positive flow rate means that water is flowing from the inlet to the outlet; a negative
 * flow rate means water is flowing from the outlet to the inlet.
 * 
 * The <code>Pipe</code> will be drawn in one of two colors, depending on whether its flow 
 * rate is positive or negative. You can get and set these with the methods 
 * <code>setColor()</code> and <code>setNegativeColor()</code>. If the flow rate is zero, 
 * the pipe is not drawn.
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Pipe extends FluidContainer {
	private double flowRate;
	private Tank inlet, outlet;
	
	/**
	 * Create a new <code>Pipe</code>, loading the image of the given <code>name</code>
	 * from the model belonging to <code>background</code>. It will connect the given
	 * <code>inlet</code> and <code>outlet</code>, either of which may be <code>null</code>
	 * to specify that the pipe is attached to something outside the simulation. The 
	 * dimensions of the pipe will be set to reflect the area of the image where this
	 * pipe is drawn, and contains() will return true only for points within the pipe.
	 * @param name Name of .png file to load
	 * @param inlet <code>Tank</code> the pipe feeds from, or <code>null</code> if it is connected to
	 *        some outside source.
	 * @param outlet <code>Tank</code> the pipe drains into, or <code>null</code> if it
	 *        leads outside of the simulation
	 * @param background <code>Background</code> to add this pipe to
	 */
	public Pipe(String name, Tank inlet, Tank outlet, Background background) {
		super(name, background);
		this.inlet = inlet;
		this.outlet = outlet;
		flowRate = 0;
	}
	
	void draw(int[] pixels, int w, int h) {
		if(flowRate == 0 || isHidden()) return;
		
		int i = 0, minY = getY(), maxY = minY + getHeight() - 1, minX = getX();
		int rgb = positiveColor;
		if(flowRate < 0) rgb = negativeColor;
	
		for(int y = minY; y <= maxY; y++) {
			int d = y * w + minX;
			for(int x = getWidth(); x > 0; x--, d++, i++) {
				pixels[d] = (pixels[d] & ~mask[i]) | (rgb & mask[i]);
			}
		}
	}
	
	/**
	 * Set the flow rate.
	 * @param rate New flow rate, in gallons per second.
	 */
	public void setFlowRate(double rate) {
		flowRate = rate;
		changed();
	}
	
	/**
	 * Get the flow rate.
	 * @return The flow rate, in gallons per second.
	 */
	public double getFlowRate() {
		return flowRate;
	}
	
	/**
	 * Get the <code>Tank</code> this pipe feeds from.
	 * @return the inlet <code>Tank</code>
	 */
	public Tank getInlet() {
		return inlet;
	}
	
	/**
	 * Get the <code>Tank</code> this pipe drains into.
	 * @return the outlet <code>Tank</code>
	 */
	public Tank getOutlet() {
		return outlet;
	}
}
