package legos;
import java.awt.Color;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;

import objectdraw.*;

/**
 * An <code>Arm</code> is a <code>LegoPiece</code> representing the arm of a lego person. It
 * has a single subunit, a <code>Hand</code>.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Arm extends LegoPiece {
	static {
		register();
	}
	
	private static float sin(double arg) {
		return (float)Math.sin(arg * Math.PI / 180.);
	}
	
	private static float cos(double arg) {
		return (float)Math.cos(arg * Math.PI / 180.);
	}
	
	private static void register() {
		float a = 25, b = 25, theta = 10;
		
		float ax = -a * sin(theta), ay = a * cos(theta);
		float bx = b * sin(60 - theta), by = b * cos(60 - theta);
		
		float dx = ax - 10;
		float dy = -10;
		AffineTransform translate = AffineTransform.getTranslateInstance(dx, dy);
		
		GeneralPath path = new GeneralPath();
		path.append(new Arc2D.Double(0, ay, 20, 20, 180 - theta + 60, -60, Arc2D.OPEN), false);
		path.append(new Arc2D.Double(-ax, 0, 20, 20, 180 - theta, -180, Arc2D.OPEN), true);
		path.lineTo(21, ay + 6);
		
		AffineTransform t = new AffineTransform();
		t.translate(bx + 10, ay + by + 10);
		t.rotate(-Math.PI * (60 - theta) / 180.);
		path.append(t.createTransformedShape(new Arc2D.Double(-10, -6, 20, 12, 0, -180, Arc2D.OPEN)), true);
		path.closePath();
		
		Shape fill = translate.createTransformedShape(path);

		path.append(t.createTransformedShape(new Arc2D.Double(-10, -6, 20, 12, 180, -180, Arc2D.OPEN)), false);
		
		path.append(t.createTransformedShape(new Ellipse2D.Double(-5, -3, 10, 6)), false);
		
		Shape outline = translate.createTransformedShape(path);
		
		Attachment[] attachments = {
				new Attachment(bx + 10 + dx, ay + by + 10 + dy, true, Hand.class),
		};
		LegoPiece.registerShape(Arm.class, Color.RED, outline, fill, attachments);
	}
	
	/**
	 * Get the <code>Hand</code>.
	 * @return the <code>Hand</code>.
	 */
	public Hand getHand() {
		return (Hand)getSubunit(0);
	}
	
	/**
	 * Replace the <code>Hand</code> with a different <code>Hand</code>.
	 * @param hand new <code>Hand</code> to replace the old one with.
	 */
	public void setHand(Hand hand) {
		setSubunit(0, hand);
	}
	
	/**
	 * Construct a handless arm, placing it on the given canvas. The sleeve will be colored red.
	 * @param canvas Canvas to place this arm on.
	 */
	public Arm(DrawingCanvas canvas) {
		this(null, null, canvas);
	}
	
	/**
	 * Construct a handless arm, placing it on the given canvas.
	 * @param color The color of the sleeve.
	 * @param canvas Canvas to place this arm on.
	 */
	public Arm(Color color, DrawingCanvas canvas) {
		this(null, color, canvas);
	}
	
	/**
	 * Construct an arm, placing it on the given canvas. The sleeve will be colored red.
	 * @param hand The <code>Hand</code>.
	 * @param canvas Canvas to place this arm on.
	 */
	public Arm(Hand hand, DrawingCanvas canvas) {
		this(hand, null, canvas);
	}
	
	/**
	 * Construct an arm, placing it on the given canvas.
	 * @param hand The <code>Hand</code>.
	 * @param color The color of the sleeve.
	 * @param canvas Canvas to place this arm on.
	 */
	public Arm(Hand hand, Color color, DrawingCanvas canvas) {
		super(new LegoPiece[] { hand }, color, canvas);
	}
	
	/**
	 * Construct a handless arm, but do not make it appear on any canvas
	 * yet. The sleeve will be colored red. You can make it appear by adding it to a canvas, or 
	 * setting it as an arm of of a <code>Body</code> that is already visible on a canvas.
	 */
	public Arm() {
		this(null, null, null);
	}
	
	/**
	 * Construct a handless arm, but do not make it appear on any canvas
	 * yet. You can make it appear by adding it to a canvas, or 
	 * setting it as an arm of of a <code>Body</code> that is already visible on a canvas.
	 * @param color The color of the sleeve.
	 */
	public Arm(Color color) {
		this(null, color, null);
	}
	
	/**
	 * Construct an arm, but do not make it appear on any canvas
	 * yet. The sleeve will be colored red. You can make it appear by adding it to a canvas, or 
	 * setting it as an arm of of a <code>Body</code> that is already visible on a canvas.
	 * @param hand The <code>Hand</code>.
	 */
	public Arm(Hand hand) {
		this(hand, null, null);
	}
	
	/**
	 * Construct an arm, but do not make it appear on any canvas
	 * yet. You can make it appear by adding it to a canvas, or 
	 * setting it as an arm of of a <code>Body</code> that is already visible on a canvas.
	 * @param hand The <code>Hand</code>.
	 * @param color The color of the sleeve.
	 */
	public Arm(Hand hand, Color color) {
		this(hand, color, null);
	}
}
