package legos;
import java.awt.Color;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;

import objectdraw.*;

/**
 * A <code>Body</code> is a <code>LegoPiece</code> representing the body of a lego person. It
 * has four subunits:<ul><li>A <code>Head</code>
 * <li>A set of <code>Legs</code>
 * <li>A right <code>Arm</code>
 * <li>A left <code>Arm</code></ul>
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Body extends LegoPiece {
	static {
		LegoPiece.registerShape(Body.class, Color.RED, buildOutline(), buildFill(), new Attachment[] {
			new Attachment(0, -20, true, Head.class),
			new Attachment(0, 40, false, Legs.class),
			new Attachment(-18.5, -10, true, Arm.class),
			new Attachment(18.5, -10, false, Arm.class),
		});
	}
	
	private static Shape buildFill() {
		GeneralPath path = new GeneralPath();
		path.moveTo(0, 60);
		path.lineTo(6, 0);
		path.lineTo(22, 0);
		path.lineTo(22, -20);
		path.lineTo(36, -20);
		path.lineTo(36, 0);
		path.lineTo(52, 0);
		path.lineTo(58, 60);
		path.closePath();
		return AffineTransform.getTranslateInstance(-29, -20).createTransformedShape(path);
	}
	
	private static Shape buildOutline() {
		GeneralPath path = new GeneralPath();
		path.moveTo(0, 60);
		path.lineTo(6, 0);
		path.lineTo(52, 0);
		path.lineTo(58, 60);
		path.closePath();
		
		path.moveTo(18, 0);
		path.lineTo(12, 60);
		
		path.moveTo(22, 0);
		path.lineTo(22, -20);
		path.lineTo(36, -20);
		path.lineTo(36, 0);
		
		path.append(new Ellipse2D.Double(9, 5, 3, 10), false);
		return AffineTransform.getTranslateInstance(-29, -20).createTransformedShape(path);
	}
	
	/**
	 * Return the <code>Head</code>.
	 * @return the <code>Head</code>.
	 */
	public Head getHead() {
		return (Head)getSubunit(0);
	}
	
	/**
	 * Return the <code>Head</code>.
	 * @return the <code>Head</code>.
	 */
	public Legs getLegs() {
		return (Legs)getSubunit(1);
	}
	
	/**
	 * Return the right <code>Arm</code>.
	 * @return the right <code>Arm</code>.
	 */
	public Arm getRightArm() {
		return (Arm)getSubunit(2);
	}
	
	/**
	 * Return the left <code>Arm</code>.
	 * @return the left <code>Arm</code>.
	 */
	public Arm getLeftArm() {
		return (Arm)getSubunit(3);
	}
	
	/**
	 * Replace the <code>Head</code> with a different <code>Head</code>.
	 * @param head new <code>Head</code> to replace the old one with.
	 */
	public void setHead(Head head) {
		setSubunit(0, head);
	}
	
	/**
	 * Replace the <code>Legs</code> with a different <code>Legs</code>.
	 * @param legs new <code>Legs</code> to replace the old one with.
	 */
	public void setLegs(Legs legs) {
		setSubunit(1, legs);
	}
	
	/**
	 * Replace the right <code>Arm</code> with a different <code>Arm</code>.
	 * @param arm new <code>Arm</code> to replace the old one with.
	 */
	public void setRightArm(Arm arm) {
		setSubunit(2, arm);
	}
	
	/**
	 * Replace the left <code>Arm</code> with a different <code>Arm</code>.
	 * @param arm new <code>Arm</code> to replace the old one with.
	 */
	public void setLeftArm(Arm arm) {
		setSubunit(3, arm);
	}
	
	/**
	 * Construct just the torso of a lego person, placing it on the given canvas. The torso will
	 * be colored red.
	 * @param canvas the canvas to add this person to.
	 */
	public Body(DrawingCanvas canvas) {
		this(null, null, null, null, null, canvas);
	}
	
	/**
	 * Construct just the torso of a lego person, placing it on the given canvas.
	 * @param color the color of the torso.
	 * @param canvas the canvas to add this person to.
	 */
	public Body(Color color, DrawingCanvas canvas) {
		this(null, null, null, null, color, canvas);
	}
	
	/**
	 * Construct a lego person, placing it on the given canvas. The torso of the person will
	 * be colored red.
	 * @param head the person's <code>Head</code>.
	 * @param legs the person's <code>Legs</code>.
	 * @param rightArm the person's right <code>Arm</code>.
	 * @param leftArm the person's left <code>Arm</code>.
	 * @param canvas the canvas to add this person to.
	 */
	public Body(Head head, Legs legs, Arm rightArm, Arm leftArm, DrawingCanvas canvas) {
		this(head, legs, rightArm, leftArm, null, canvas);
	}
	
	/**
	 * Construct a lego person, placing it on the given canvas.
	 * @param head the person's <code>Head</code>.
	 * @param legs the person's <code>Legs</code>.
	 * @param rightArm the person's right <code>Arm</code>.
	 * @param leftArm the person's left <code>Arm</code>.
	 * @param color the color of the person's torso.
	 * @param canvas the canvas to add this person to.
	 */
	public Body(Head head, Legs legs, Arm rightArm, Arm leftArm, Color color, DrawingCanvas canvas) {
		super(new LegoPiece[] { head, legs, rightArm, leftArm }, color, canvas);
	}
}
