package legos;
import objectdraw.*;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.util.HashMap;

/**
 * A <code>LegoPiece</code> represents some kind of object created by connecting together smaller
 * pieces. Each <code>LegoPiece</code> has a shape of its own, but is also able to have more
 * <code>LegoPiece</code>s of specific types attached to it at particular points. So, for example,
 * a <code>LegoPiece</code> representing the body of a lego man only draws the body when it is by
 * itself, but can have a head, legs, and arms attached to it.
 * 
 * The <code>LegoPiece</code> class contains methods for accessing the attached pieces (subunits)
 * by simply specifying the number of the subunit. More often you will use methods in a particular
 * subclass that retrieve a particular subunit. So, for example, I could get the head of a lego person
 * with <code>body.getSubunit(0)</code>, but it would make more sense to use the <code>getHead()</code>
 * method in the <code>Body</code> class.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 *
 */
public abstract class LegoPiece extends ObjectDrawShape implements LocatableInterface {
	private Shape outline;
	private Shape fill;
	private LegoPiece[] subunits;
	private LegoPiece owner;
	
	private static final boolean SHOW_ATTACHMENTS = false;
	
	private static HashMap shapes = new HashMap();
	
	private static class Shapes {
		public Shape outline;
		public Shape fill;
		public Color defaultColor;
		public Attachment[] attachments;
		
		public Shapes(Shape outline, Shape fill, Color defaultColor, Attachment[] attachments) {
			this.outline = outline;
			this.fill = fill;
			this.defaultColor = defaultColor;
			this.attachments = attachments;
		}
	}
	
	static void registerShape(Class c, Color defaultColor, Shape outline, Shape fill) {
		registerShape(c, defaultColor, outline, fill, new Attachment[0]);
	}
	
	static void registerShape(Class c, Color defaultColor, Shape outline, Shape fill, Attachment[] attachments) {
		if(shapes.containsKey(c)) return;
		
		shapes.put(c, new Shapes(outline, fill, defaultColor, attachments));
	}
	
	private static final Shape EMPTY_SHAPE = new GeneralPath();
	private static final Shapes EMPTY_SHAPES = new Shapes(EMPTY_SHAPE, EMPTY_SHAPE, Color.BLUE, new Attachment[0]);
	private static final BasicStroke STROKE = new BasicStroke(2f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
	
	private static String className(Class c) {
		String name = c.getName();
		return name.substring(name.lastIndexOf(".") + 1);
	}
	
	LegoPiece(Color color, DrawingCanvas canvas) {
		this(new LegoPiece[0], color, canvas);
	}
	
	LegoPiece(LegoPiece[] pieces, Color color, DrawingCanvas canvas) {
		super(new Location(0, 0), true, color, canvas);
		
		Attachment[] attachments = getShapes().attachments;
		if(attachments.length != pieces.length) {
			throw new IllegalArgumentException("Number of pieces (" + pieces.length + 
					") does not match number of attachments (" + attachments.length + ").");
		}
		
		owner = null;
		subunits = new LegoPiece[pieces.length];
		for(int i = 0; i < subunits.length; i++) {
			setSubunit(i, pieces[i]);
		}
		
		if(color == null) setColor(getShapes().defaultColor);
		
		if(canvas != null) {
			Rectangle2D bounds = getShape().getBounds2D();
			moveTo(-bounds.getX() + Math.random() * (canvas.getWidth() - bounds.getWidth()),
					-bounds.getY() + Math.random() * (canvas.getHeight() - bounds.getHeight()));
		}
		
		ready();
	}
	
	private void checkIndex(int index) {
		if(index < 0) {
			throw new IndexOutOfBoundsException("The index of a subunit cannot be less than zero. You specified an index of " + index + ".");
		}
		Attachment[] attachments = getShapes().attachments;
		if(attachments.length == 0) {
			throw new IndexOutOfBoundsException("A " + className(getClass()) + " cannot have any subunits.");
		} else if(index >= attachments.length) {
			throw new IndexOutOfBoundsException("A " + className(getClass()) + " has only " + attachments.length + " subunits, but you specified an index of " + index + ". The index must be between 0 and " + (attachments.length - 1) + ".");
		}
	}
	
	/**
	 * Return the number of subunits that can be attached to this piece. So, for example, a <code>Legs</code>
	 * will return 2, because there are two places where a <code>Leg</code> can be attached to it.
	 * @return the number of subunits that can be attached to this piece.
	 */
	public int getSubunitCount() {
		return subunits.length;
	}
	
	/**
	 * Retrieve one of the subunits of this piece
	 * @param index Index of subunit, between <code>0</code> and <code>getSubunitCount() - 1</code>.
	 * @return The <code>LegoPiece</code> at that index, or <code>null</code> if there is nothing attached there.
	 */
	public LegoPiece getSubunit(int index) {
		checkIndex(index);
		return subunits[index];
	}
	
	/**
	 * Return the point at which the given subunit attaches. This will give you the same location that you
	 * would get with <code>getSubunit(index).getLocation()<code>, except that it will return a valid
	 * <code>Location</code> even if there is no subunit currently attached there.
	 * @param index Index of subunit, between <code>0</code> and <code>getSubunitCount() - 1</code>.
	 * @return <code>Location</code> at which the given subunit would be placed if it were attached.
	 */
	public Location getAttachmentPoint(int index) {
		checkIndex(index);
		Attachment attachment = getShapes().attachments[index];
		return new Location(getDoubleX() + attachment.x, getDoubleY() + attachment.y);
	}
	
	/**
	 * Check whether the given piece can be attached at the given index. If this returns <code>true</code>,
	 * you can safely call <code>setSubunit(index, subunit)</code> with the same index and subunit; if it
	 * returns <code>false</code>, attaching that subunit would be illegal.
	 * @param index Index to attach subunit at, between <code>0</code> and <code>getSubunitCount() - 1</code>.
	 * @param subunit Subunit that will be checked for whether it can be attached there.
	 * @return <code>true</code> if the given piece can be attached at the given index, <code>false</code> otherwise.
	 */
	public boolean canAttach(int index, LegoPiece subunit) {
		checkIndex(index);
		if(subunit == null) return true;
		
		Attachment attachment = getShapes().attachments[index];
		return attachment.type.isAssignableFrom(subunit.getClass());
	}
	
	public LegoPiece getOwner() {
		return owner;
	}
	
	/**
	 * Return an index at which the given subunit could be attached to this piece. If there are no attachment
	 * points on this piece that match the type of the given subunit, this method will return -1. If there is
	 * more than one valid attachment point, the index of the one closer to the location of the given subunit
	 * will be returned.
	 * @param subunit Subunit for which an attachment point is sought
	 * @return Index of closest valid attachment point, or -1 if there are no valid attachment points.
	 */
	/*public int getAttachmentIndex(LegoPiece subunit) {
		int closest = -1;
		double minDistance = Double.POSITIVE_INFINITY;
		Location location = subunit.getLocation();
		
		for(int i = getSubunitCount() - 1; i >= 0; i--) {
			if(canAttach(i, subunit)) {
				double distance = location.distanceTo(getAttachmentPoint(i));
				if(distance < minDistance) {
					minDistance = distance;
					closest = i;
				}
			}
		}
		return closest;
	}*/
	
	/**
	 * Attach the given piece at the given index. If there is already a piece there, it will be removed from
	 * the canvas. If the piece being added is already a subunit of another piece, it will be detached from
	 * that piece before being attached to this one. The piece to be attached will also be automatically moved
	 * to the correct position. An error will occur if the piece is not the proper type to be attached at the
	 * given index.
	 * @param index Index to attach subunit at, between <code>0</code> and <code>getSubunitCount() - 1</code>.
	 * @param subunit Subunit that will be attached there.
	 */
	public void setSubunit(int index, LegoPiece subunit) {
		if(!canAttach(index, subunit)) {
			throw new ClassCastException("This piece requires a " + className(getShapes().attachments[index].type)
					+ " at index " + index + "; you gave me a " + className(subunit.getClass()) + ".");
		}
		
		if(subunit == null) { 
			removeSubunit(index);
		} else {
			if(subunit.owner != null) subunit.owner.removeSubunit(subunit);
			else if(subunit.getCanvas() != null) subunit.removeFromCanvas();
			
			if(subunits[index] != null) removeSubunit(index);
			
			subunits[index] = subunit;
			subunit.owner = this;
			subunit.depend(this);
			
			update();
		}
	}
	
	/**
	 * Remove the given subunit from this piece.
	 * @param index of subunit o remove
	 */
	public void removeSubunit(int index) {
		checkIndex(index);
		
		if(subunits[index] == null) return;
		
		subunits[index].owner = null;
		subunits[index].undepend(this);
		subunits[index] = null;
		
		update();
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void update() {
		super.update();
		for(int i = 0; i < subunits.length; i++) {
			if(subunits[i] != null) subunits[i].update();
		}
	}
	
	/**
	 * Remove the given subunit from this piece.
	 * @param subunit Piece to remove
	 */
	public void removeSubunit(LegoPiece subunit) {
		for(int i = 0; i < subunits.length; i++) {
			if(subunits[i] == subunit) {
				removeSubunit(i);
				return;
			}
		}
	}
	
	/**
	 * Remove this piece from the piece it is attached to. It will appear on the canvas a short distance away.
	 */
	public void detach() {
		if(owner == null) return;
		
		LegoPiece oldOwner = owner;
		owner.removeSubunit(this);
		
		addToCanvas(oldOwner.getCanvas());
		
		// Move the subunit out 20 pixels from this unit
		double dx = getDoubleX() - oldOwner.getDoubleX();
		double dy = getDoubleY() - oldOwner.getDoubleY();
		double d = 45 / Math.sqrt(dx * dx + dy * dy);
		
		move(dx * d, dy * d);
	}
	
	private Shapes getShapes() {
		Class c = getClass();
		Shapes ss = null;
		
		while(ss == null) {
			ss = (Shapes)shapes.get(c);
			c = c.getSuperclass();
			if(c == Object.class) {
				ss = EMPTY_SHAPES;
			}
		}
		return ss;
	}

	private Shape getShape(boolean outline) {
		Shapes ss = getShapes();
		
		Shape s = (outline ? ss.outline : ss.fill);
		
		return AffineTransform.getTranslateInstance(getDoubleX(), getDoubleY()).createTransformedShape(s);
	}
	
	private Shape getOutline() {
		return getShape(true);
	}
	
	private Shape getFill() {
		return getShape(false);
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public Shape makeShape() {
		return makeArea();
	}
	
	private Area makeArea() {
		Area fullShape = new Area();
		
		fill = getFill();
		outline = getOutline();
		fullShape.add(new Area(fill));
		
		for(int i = 0; i < subunits.length; i++) {
			if(subunits[i] != null) {
				subunits[i].moveTo(getAttachmentPoint(i));
				fullShape.add(subunits[i].makeArea());
			}
		}
		return fullShape;
	}
	
	private Shape translate(Shape shape, Location location) {
		return AffineTransform.getTranslateInstance(location.getDoubleX(), location.getDoubleY())
					.createTransformedShape(shape);
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void draw(Graphics2D g) {
		if(isHidden()) return;
		
		getShape();
		
		Attachment[] attachments = getShapes().attachments;
		for(int i = 0; i < subunits.length; i++) {
			if(!attachments[i].above && subunits[i] != null) {
				subunits[i].draw(g);
			}
		}
		
		g.setColor(getColor());
		g.fill(fill);
		g.setColor(Color.BLACK);
		g.setStroke(STROKE);
		g.draw(outline);
		
		if(SHOW_ATTACHMENTS) {
			g.setColor(Color.GRAY);
			GeneralPath mark = new GeneralPath();
			mark.moveTo(-3, -3);
			mark.lineTo(3, 3);
			mark.moveTo(-3, 3);
			mark.lineTo(3, -3);
			g.draw(translate(mark, getLocation()));
			for(int i = getSubunitCount() - 1; i >= 0; i--) {
				g.draw(translate(mark, getAttachmentPoint(i)));
			}
		}
		
		for(int i = 0; i < subunits.length; i++) {
			if(attachments[i].above && subunits[i] != null) {
				subunits[i].draw(g);
			}
		}
	}

	/* * * * Methods from LocatableInterface * * * */
	
	/**
	 * Get this object's x coordinate.
	 * @return this object's x coordinate
	 */
	public int getX() { return getLocation().getX(); }
	/**
	 * Get this object's y coordinate.
	 * @return this object's y coordinate
	 */
	public int getY() { return getLocation().getY(); }
	/**
	 * Get this object's x coordinate, in double precision.
	 * @return this object's x coordinate, in double precision
	 */
	public double getDoubleX() { return getLocation().getDoubleX(); }
	/**
	 * Get this object's y coordinate, in double precision.
	 * @return this object's y coordinate, in double precision
	 */
	public double getDoubleY() { return getLocation().getDoubleY(); }
	/**
	 * Get this object's location.
	 * @return the <code>Location</code> of this object.
	 */
	public Location getLocation() { return getMyLocation(); }

	/* * * * Methods from DrawableInterface * * * */
	
	/**
	 * Make this object invisible. It stays in position on its canvas; it is simply not drawn.
	 * This method has no effect if the object was already hidden.
	 */
	public void hide() { super.hide(); }
	/**
	 * Make an object visible again after it had been hidden. This method has no effect if the
	 * object was already visible.
	 */
	public void show() { super.show(); }
	/**
	 * Return <code>true</code> if this object has been rendered invisible with a call to its
	 * <code>hide()</code> method, <code>false</code> if it is still being drawn.
	 * @return whether this object is hidden
	 */
	public boolean isHidden() { return super.isHidden(); }

	/**
	 * Place this object on the specified canvas. The object will be removed from its current
	 * canvas, if it is on a canvas that is not the one it is being moved to. The object will
	 * be at the front of the canvas, as if it had just been drawn. If the given canvas is 
	 * <code>null</code> the object is simply removed from its current canvas.
	 * @param c <code>DrawingCanvas</code> to place this object on
	 */
	public void addToCanvas(DrawingCanvas c) { super.addToCanvas(c); }
	/**
	 * Remove this object from its canvas.
	 */
	public void removeFromCanvas() { super.removeFromCanvas(); }
	/**
	 * Return the canvas that this object is on. If it has been removed from its canvas, this
	 * method will return <code>null</code>.
	 * @return the <code>DrawingCanvas</code> that this object is on.
	 */
	public DrawingCanvas getCanvas() { return super.getCanvas(); }

	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param point <code>Location</code> to move to
	 */
	public void moveTo(Location point) { super.moveTo(point); }
	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param x x coordinate to move to
	 * @param y y coordinate to move to
	 */
	public void moveTo(double x, double y) { super.moveTo(x, y); }
	/**
	 * Shift (translate) this object left or right by an amount <code>dx</code>, and up or
	 * down by an amount <code>dy</code>. A negative <code>dx</code> indicates a shift to the
	 * left; a negative <code>dy</code> indicates a shift up
	 * @param dx amount of x translation
	 * @param dy amount of y translation
	 */
	public void move(double dx, double dy) { super.move(dx, dy); }

	/**
	 * Return the color of this object
	 * @return the <code>Color</code> of this object
	 */
	public Color getColor() { return super.getColor(); }
	/**
	 * Change the color of this object.
	 * @param c new <code>Color</code> for this object
	 */
	public void setColor(Color c) { super.setColor(c); }

	/**
	 * Move this object one step forward in the draw order, causing it to be drawn on top of
	 * the objects that are behind it.
	 */
	public void sendForward() {
		if(owner == null) { super.sendForward(); } else { owner.sendForward(); }
	}
	/**
	 * Move this object one step backward in the draw order, causing it to be drawn underneath
	 * the objects that are in front of it.
	 */
	public void sendBackward() {
		if(owner == null) { super.sendBackward(); } else { owner.sendBackward(); }
	}
	/**
	 * Move this object to the very front of the draw order, causing it to be drawn on top of
	 * all the other objects.
	 */
	public void sendToFront() {
		if(owner == null) { super.sendToFront(); } else { owner.sendToFront(); }
	}
	/**
	 * Move this object to the very back of the draw order, causing it to be drawn underneath
	 * all the other objects.
	 */
	public void sendToBack() {
		if(owner == null) { super.sendToBack(); } else { owner.sendToBack(); }
	}

	/**
	 * Return <code>true</code> if the given location is inside this object, <code>false</code>
	 * otherwise. Because <code>Line</code>s and <code>FramedArc</code>s have no "inside", a
	 * point is considered to be contained in one of them as long as it is within a few pixels
	 * of the line.
	 * @param point <code>Location</code> to test
	 * @return whether the given location is inside this shape
	 */
	public boolean contains(Location point) {
		return getShape().contains(point.getDoubleX(), point.getDoubleY());
	}

	/**
	 * Return a string describing how this object might be constructed.
	 */
	public String toString() {
		String str = "";
		for(int i = 0; i < subunits.length; i++) {
			str += subunits[i];
			if(i != subunits.length - 1) str += ", ";
		}
		str = super.toString(str);
		str = str.replaceAll(",\\s+null\\)", ")");
		str = str.replaceAll("\\(,\\s+", "(");
		return str;
	}
}
