package seminar;
import java.awt.Color;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Random;

import objectdraw.*;

/**
 * <code>ObjectCreationPractice</code> is a game that someone beginning to learn ObjectDraw can use
 * to familiarize themselves with the basic object types and how to create them. The method
 * <code>ObjectCreationPractice.show()</code> displays a pattern of five objects on the screen 
 * for the programmer to try to write instructions to match. The method <code>ObjectCreationPractice.test()</code>
 * will actually verify whether the objects currently on the canvas match the desired pattern.
 * <br><br>
 * Each of these methods takes an <code>int</code> parameter specifying which pattern to show
 * or test. The pattern that you are trying to match from <code>show()</code> should be the same
 * number you ask the program to <code>test</code>. The number of the pattern also controls the
 * difficulty:<ul><li>0-999: Only <code>FramedRect</code> and <code>FilledRect</code> will be used,
 * with no color parameter.
 * <li>1000-1999: <code>FramedOval</code> and <code>FilledOval</code> are now possible, but still
 * with no color parameter.
 * <li>2000-2999: Colors are introduced. The only possibilities are <code>Color.RED</code>,
 * <code>Color.GREEN</code>, <code>Color.BLUE</code>, <code>Color.YELLOW</code>, <code>Color.CYAN</code>,
 * <code>Color.MAGENTA</code>, <code>Color.GRAY</code>, and <code>Color.BLACK</code>.
 * <li>3000-3999: <code>Line</code>s are introduced.
 * <li>4000-4999: <code>FramedRoundedRect</code> and <code>FilledRoundedRect</code> are introduced.
 * <li>5000+: <code>FramedArc</code> and <code>FilledArc</code> are introduced.
 * </ul>All coordinates will always be in multiples of 10. Arc widths for rounded rects are in multiples
 * of 20. Angles for arcs are in multiples of 30. Keeping this in mind dramatically reduces the 
 * amount of guesswork needed.
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class ObjectCreationPractice {
	private static final Color[] colors = {Color.BLACK, Color.BLUE, Color.CYAN, 
		Color.GRAY, Color.GREEN, Color.MAGENTA, Color.RED, Color.YELLOW};
	private static final int[] typesForLevel = {2, 4, 4, 5, 7, 9, 9};
	private static final DrawableComparator comparator = new DrawableComparator();
	
	private static class DrawableComparator implements Comparator {
		public int compare(final Object a, final Object b) {
			int comparison;
			
			if((comparison = compareTopLeft(a, b)) != 0) return comparison;
			else if((comparison = compareBottomRight(a, b)) != 0) return comparison;
			else if((comparison = compareClass(a, b)) != 0) return comparison;
			else if((comparison = compareColor(a, b)) != 0) return comparison;
			else return compareOther(a, b);
		}
	};
	
	private static Location getTopLeft(final Object obj) {
		if(obj instanceof Line) {
			final Line line = (Line)obj;
			return new Location(Math.min(line.getStart().getX(), line.getEnd().getX()),
					Math.min(line.getStart().getY(), line.getEnd().getY()));
		} else {
			return new Location(((Drawable2DInterface)obj).getLocation());
		}
	}
	
	private static Location getBottomRight(final Object obj) {
		if(obj instanceof Line) {
			final Line line = (Line)obj;
			return new Location(Math.max(line.getStart().getX(), line.getEnd().getX()),
					Math.max(line.getStart().getY(), line.getEnd().getY()));
		} else {
			final Drawable2DInterface box = (Drawable2DInterface)obj;
			return new Location(box.getX() + box.getWidth(), box.getY() + box.getHeight());
		}
	}
	
	private static int compareLocations(final Location loc1, final Location loc2) {
		return (loc1.getY() == loc2.getY() ? loc1.getX() - loc2.getX() : loc1.getY() - loc2.getY());
	}
	
	private static int compareTopLeft(final Object a, final Object b) {
		return compareLocations(getTopLeft(a), getTopLeft(b));
	}
	
	private static int compareBottomRight(final Object a, final Object b) {
		return compareLocations(getBottomRight(a), getBottomRight(b));
	}
	
	private static int compareClass(final Object a, final Object b) {
		return a.getClass().getName().compareTo(b.getClass().getName());
	}
	
	private static int compareColor(final Object a, final Object b) {
		return ((DrawableInterface)a).getColor().getRGB() - ((DrawableInterface)b).getColor().getRGB();
	}
	
	private static int slant(final Line a, final Line b) {
		// dx1 * dy1 - dx2 * dy2
		return (a.getEnd().getX() - a.getStart().getX()) * (b.getEnd().getY() - b.getStart().getY()) -
				(b.getEnd().getX() - b.getStart().getX()) * (a.getEnd().getY() - a.getStart().getY());
	}
	
	private static int[] arcs(Object a) {
		if(a instanceof FramedRoundedRect) {
			return new int[] { ((FramedRoundedRect)a).getArcWidth(), ((FramedRoundedRect)a).getArcHeight() };
		} else {
			return new int[] { ((FilledRoundedRect)a).getArcWidth(), ((FilledRoundedRect)a).getArcHeight() };
		}
	}
	
	private static int[] angles(Object a) {
		int[] angles = new int[2];
		
		if(a instanceof FramedArc) {
			angles[0] = (int)(((FramedArc)a).getStartAngle() + .5);
			angles[1] = (int)(((FramedArc)a).getArcAngle() + .5);
		} else {
			angles[0] = (int)(((FilledArc)a).getStartAngle() + .5);
			angles[1] = (int)(((FilledArc)a).getArcAngle() + .5);
		}
		
		// Make arc angle positive
		if(angles[1] < 0) {
			angles[0] += angles[1];
			angles[1] = -angles[1];
		}
		
		// Make start angle from 0 to 360
		while(angles[0] < 0) {
			angles[0] += 360;
		}
		if(angles[0] >= 360) angles[0] -= 360;
		
		return angles;
	}
	
	private static boolean isRect(Object a) {
		return a instanceof FramedRect || a instanceof FilledRect;
	}
	
	private static boolean isOval(Object a) {
		return a instanceof FramedOval || a instanceof FilledOval;
	}
	
	private static boolean isRoundedRect(Object a) {
		return a instanceof FramedRoundedRect || a instanceof FilledRoundedRect;
	}
	
	private static boolean isArc(Object a) {
		return a instanceof FramedArc || a instanceof FilledArc;
	}
	
	private static int compareOther(final Object a, final Object b) {
		if(a instanceof Line) {
			return slant((Line)a, (Line)b);
		} else if(isRoundedRect(a)) {
			final int[] arcs1 = arcs(a);
			final int[] arcs2 = arcs(b);
			
			return (arcs1[1] == arcs2[1] ? arcs1[0] - arcs2[0] : arcs1[1] - arcs2[1]);
		} else if(isArc(a)) {
			final int[] angles1 = angles(a);
			final int[] angles2 = angles(b);
			
			return (angles1[1] == angles2[1] ? angles1[0] - angles2[0] : angles1[1] - angles2[1]);
		} else {
			return 0;
		}
	}
	
	private static Location[] createEndpoints(final DrawingCanvas canvas, final Random random) {
		final int width = random.nextInt((canvas.getWidth() - 70) / 40) * 10 + 50;
		final int height = random.nextInt((canvas.getWidth() - 70) / 40) * 10 + 50;
		final int x1 = random.nextInt((canvas.getWidth() - width - 20) / 10) * 10 + 10;
		final int y1 = random.nextInt((canvas.getHeight() - height - 20) / 10) * 10 + 10;
		
		final Location arcWidths = new Location(
				random.nextInt((width - 30) / 20) * 20 + 20, // From 20 to width - 30, in multiples of 20
				random.nextInt((height - 30) / 20) * 20 + 20);
		
		final Location arcAngles = new Location(
				random.nextInt(12) * 30, // From 0 to 330 in multiples of 30
				random.nextInt(6) * 30 + 180); // From 180 to 330 in multiples of 30
		
		if(random.nextInt() % 2 == 0) {
			return new Location[] { new Location(x1, y1), new Location(x1 + width, y1 + height), arcWidths, arcAngles };
		} else {
			return new Location[] { new Location(x1 + width, y1), new Location(x1, y1 + height), arcWidths, arcAngles };
		}
	}

	private static DrawableInterface createObject(final DrawingCanvas canvas, final Random random, final int level) {
		// Level 0: FramedRect, FilledRect, FramedOval, and FilledOval, all Color.BLACK.
		// Level 1: Any of the pre-defined colors, except white
		// Level 2: Lines are introduced
		// Level 3: FilledRoundRect and FramedRoundRect are introduced
		// Level 4: FilledArc and FramedArc are introduced
		// Level 5: Overlap is allowed and checked for
		final Color color = (level <= 1 ? Color.BLACK : colors[random.nextInt(colors.length)]);
		final Location[] points = createEndpoints(canvas, random);
		
		switch(random.nextInt(typesForLevel[level])) {
		case 0:
			return new FramedRect(points[0], points[1], color, null);
		case 1:
			return new FilledRect(points[0], points[1], color, null);
		case 2:
			return new FramedOval(points[0], points[1], color, null);
		case 3:
			return new FilledOval(points[0], points[1], color, null);
		case 4:
			return new Line(points[0], points[1], color, null);
		case 5:
			return new FramedRoundedRect(points[0], points[1], points[2].getX(), points[2].getY(), color, null);
		case 6:
			return new FilledRoundedRect(points[0], points[1], points[2].getX(), points[2].getY(), color, null);
		case 7:
			return new FramedArc(points[0], points[1], points[3].getX(), points[3].getY(), color, null);
		default:
			return new FilledArc(points[0], points[1], points[3].getX(), points[3].getY(), color, null);
		}
	}
	
	private static boolean overlaps(DrawableInterface a, DrawableInterface b) {
		if(a instanceof Line) a = new FramedRect(((Line)a).getStart(), ((Line)a).getEnd(), null);
		if(b instanceof Line) b = new FramedRect(((Line)b).getStart(), ((Line)b).getEnd(), null);
		
		return ((Drawable2DInterface)a).overlaps((Drawable2DInterface)b);
	}
	
	/**
	 * Display a pattern of five objects that you can then attempt to write code to match.
	 * @param n pattern to display
	 * @param canvas canvas to draw the pattern on
	 */
	public static void show(final int n, final DrawingCanvas canvas) {
		addAll(generate(n, canvas), canvas);
	}
	
	private static void addAll(ArrayList list, DrawingCanvas canvas) {
		for(int i = 0; i < list.size(); i++) {
			((DrawableInterface)list.get(i)).addToCanvas(canvas);
		}
	}
	
	/**
	 * Test to see whether the objects on the given canvas match a particular test pattern.
	 * Objects will be marked in one of three ways:<ul>
	 * <li>If an object matches one in the test pattern, it will be outlined in green.
	 * <li>If an object that appears on the canvas is not in the test pattern, it will be outlined
	 * in blue.
	 * <li>If an object in the test pattern does not appear on the canvas, it will appear
	 * flashing and outlined in red.
	 * </ul>The test objects that do not match are made to flash in and out of transparency to make
	 * it easier to see through them to the objects that you had actually placed. Note that even a
	 * very small difference between an object on the canvas and one in the test set will result in
	 * the one on the canvas being outlined in blue and the one in the test set appearing flashing
	 * and in red.
	 * @param n Pattern that you are attempting to match - the same number you would pass to <code>show()</code>
	 * @param canvas canvas which should be tested to see if its objects match the test set.
	 */
	public static void test(final int n, final DrawingCanvas canvas) {
		// Create a list of objects to create, and sort them by y, then x, so
		// that we can easily compare to the list of objects existing on the
		// canvas. Discard generated objects that overlap existing ones.
		final ArrayList testObjects = generate(n, canvas);
		
		final ArrayList existingObjects = new ArrayList();
		for(final DrawableIterator iter = canvas.getDrawableIterator(); iter.hasNext(); ) {
			final DrawableInterface object = iter.next();
			
			if(object instanceof Text || object instanceof VisibleImage || object instanceof Turtle) continue;
			
			existingObjects.add(object);
		}
		
		Collections.sort(testObjects, comparator);
		Collections.sort(existingObjects, comparator);
		
		final ArrayList greenLight = new ArrayList();
		final ArrayList redLight = new ArrayList();
		final ArrayList blueLight = new ArrayList();
		
		int a = 0;
		int b = 0;
		
		while(a < testObjects.size() && b < existingObjects.size()) {
			final Object test = testObjects.get(a);
			final Object existing = existingObjects.get(b);
			final int comparison = comparator.compare(test, existing);
			
			if(comparison == 0) {
				greenLight.add(existing);
				a++;
				b++;
			} else if(comparison < 0) {
				// a is smaller than b -> test object was not found
				redLight.add(test);
				a++;
			} else {
				// a is larger than b -> existing object is extraneous
				blueLight.add(existing);
				b++;
			}
		}
		
		while(a < testObjects.size()) redLight.add(testObjects.get(a++));
		while(b < existingObjects.size()) blueLight.add(existingObjects.get(b++));
		
		final ArrayList pulseList = new ArrayList(redLight);
		makeHalos(greenLight, Color.GREEN, canvas);
		pulseList.addAll(makeHalos(blueLight, Color.BLUE, canvas));
		addAll(redLight, canvas);
		pulseList.addAll(makeHalos(redLight, Color.RED, canvas));
		
		new EventTimer(new Pulsar(pulseList), "pulse").start(.05, .05);
	}
	
	public static class Pulsar {
		private final ArrayList pulseList;
		private int phase;
		
		private static final int PULSE_LENGTH = 30;
		
		public Pulsar(final ArrayList pulseList) {
			this.pulseList = pulseList;
		}
		
		public void pulse() {
			if(++phase >= PULSE_LENGTH) phase = 0;
			
			final int alpha = (int)(192 + 63 * Math.cos(phase * 2 * Math.PI / PULSE_LENGTH)) << 24;
			for(int i = 0; i < pulseList.size(); i++) {
				final DrawableInterface object = (DrawableInterface)pulseList.get(i);
				object.setColor(new Color((object.getColor().getRGB() & 0xFFFFFF) | alpha, true));
			}
		}
	}
	
	private static ArrayList generate(final int n, final DrawingCanvas canvas) {
		final Random random = new Random(n);
		final int level = (n >= 5000 ? 5 : (n < 0 ? 0 : n / 1000));
		
		final ArrayList testObjects = new ArrayList();
		GenerateLoop: while(testObjects.size() < 5) {
			final DrawableInterface object = createObject(canvas, random, level);
			for(int i = 0; i < testObjects.size(); i++) {
				if(overlaps(object, (DrawableInterface)testObjects.get(i))) continue GenerateLoop;
			}
			testObjects.add(object);
		}
		
		return testObjects;
	}
	
	private static ArrayList makeHalos(final ArrayList objects, final Color color, final DrawingCanvas canvas) {
		final ArrayList halos = new ArrayList();
		
		for(int i = 0; i < objects.size(); i++) makeHalo(objects.get(i), color, halos, canvas);
		
		return halos;
	}
	
	private static void makeHalo(final Object object, final Color color, final ArrayList list, final DrawingCanvas canvas) {
		if(object instanceof Line) {
			makeLineHalo((Line)object, color, list, canvas);
		} else {
			final Location topLeft = getTopLeft(object);
			final Location bottomRight = getBottomRight(object);
			
			topLeft.translate(-5, -5);
			bottomRight.translate(5, 5);
			
			if(isRect(object)) {
				list.add(new FramedRoundedRect(topLeft, bottomRight, 10, 10, color, canvas));
			} else if(isOval(object)) {
				list.add(new FramedOval(topLeft, bottomRight, color, canvas));
			} else if(isRoundedRect(object)) {
				// Rounded rect
				int[] arcs = arcs(object);
				
				list.add(new FramedRoundedRect(topLeft, bottomRight, arcs[0] + 10, arcs[1] + 10, color, canvas));
			} else {
				// Arcs
				final int[] angles = angles(object);
				
				list.add(new FramedArc(topLeft, bottomRight, angles[0], angles[1], color, canvas));
				
				final double cx = .5 * (topLeft.getX() + bottomRight.getX());
				final double cy = .5 * (topLeft.getY() + bottomRight.getY());
				
				final double rx = bottomRight.getX() - 5 - cx;
				final double ry = bottomRight.getY() - 5 - cy;
				
				final double ang1 = -angles[0] * Math.PI / 180;
				final double ang2 = ang1 - angles[1] * Math.PI / 180;
				
				final double x1 = cx + rx * Math.cos(ang1);
				final double y1 = cy + ry * Math.sin(ang1);
				final double x2 = cx + rx * Math.cos(ang2);
				final double y2 = cy + ry * Math.sin(ang2);
				
				final int cornerAngle;
				if(object instanceof FramedArc) {
					topLeft.translate(10, 10);
					bottomRight.translate(-10, -10);
					list.add(new FramedArc(topLeft, bottomRight, angles[0], angles[1], color, canvas));
					cornerAngle = 180;
				} else {
					final double dx1 = x1 - cx, dy1 = y1 - cy, dx2 = x2 - cx, dy2 = y2 - cy;
					final double length1 = Math.sqrt(dx1 * dx1 + dy1 * dy1);
					final double length2 = Math.sqrt(dx2 * dx2 + dy2 * dy2);
					final double ex1 = 5 * dx1 / length1, ey1 = 5 * dy1 / length1;
					final double ex2 = 5 * dx2 / length2, ey2 = 5 * dy2 / length2;
					
					// The two lines start at (cx - ey1, cy + ex1) and (cx + ey2, cy - ey1)
					// and go in the directions given by (ex1, ey1) and (ex2, ey2). Where do they intersect?
					
					if(angles[1] == 180) {
						list.add(new Line(x1 - ey1, y1 + ex1, x2 + ey2, y2 - ex2, color, canvas));
					} else {
						final double t2 = (ex1 * (ex1 + ex2) + ey1 * (ey1 + ey2)) / (ex1 * ey2 - ey1 * ex2);
						final double ax = cx + ey2 + t2 * ex2;
						final double ay = cy - ex2 + t2 * ey2;
						
						list.add(new Line(x1 - ey1, y1 + ex1, ax, ay, color, canvas));
						list.add(new Line(ax, ay, x2 + ey2, y2 - ex2, color, canvas));
					}
					cornerAngle = 90;
				}
				list.add(new FramedArc(x1 - 5, y1 - 5, 10, 10, angles[0], -cornerAngle, color, canvas));
				list.add(new FramedArc(x2 - 5, y2 - 5, 10, 10, angles[0] + angles[1], cornerAngle, color, canvas));
			}
		}
	}
	
	private static void makeLineHalo(final Line line, final Color color, final ArrayList list, final DrawingCanvas canvas) {
		final int x1 = line.getStart().getX();
		final int y1 = line.getStart().getY();
		final int x2 = line.getEnd().getX();
		final int y2 = line.getEnd().getY();
		final int dx = x2 - x1;
		final int dy = y2 - y1;
		
		final double length = Math.sqrt(dx * dx + dy * dy);
		final double ex = 5 * dx / length;
		final double ey = 5 * dy / length;
		
		final double angle = -Math.atan2(dy, dx) * 180 / Math.PI;
		
		list.add(new Line(x1 + ey, y1 - ex, x2 + ey, y2 - ex, color, canvas));
		list.add(new Line(x1 - ey, y1 + ex, x2 - ey, y2 + ex, color, canvas));
		list.add(new FramedArc(x1 - 5, y1 - 5, 10, 10, angle + 90, 180, color, canvas));
		list.add(new FramedArc(x2 - 5, y2 - 5, 10, 10, angle - 90, 180, color, canvas));
	}
}
