package objectdraw;

/**
 * The <code>ActiveObject</code> class allows you to create a separate "thread" of a program
 * that runs its own code at its own pace, independently of other code that is running. You can 
 * think of each instance of <code>ActiveObject</code> as being a worker helping to do some task;
 * there may be many workers working at the same time, each doing their own job.
 * <br><br>
 * You will use <code>ActiveObject</code> by defining a class that extends it, and writing
 * in that class a <code>run()</code> method containing the work you want that object to do:
 * <br><br><code>
 * public class NewWorkerClass extends ActiveObject {<br>
 * &nbsp;&nbsp;&nbsp;public void run() {<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;// Your code here<br>
 * &nbsp;&nbsp;&nbsp;}<br>
 * }
 * </code><br><br>
 * Once you have defined a new class of <code>ActiveObject</code>, you can then create new
 * workers and get them started on their work by calling their <code>start()</code> method:
 * <br><br><code>
 * NewWorkerClass worker = new NewWorkerClass();<br>
 * worker.start();
 * </code><br><br>
 * Typically, instead of doing all its work at once, you will will want the worker to pause after
 * each step, so that you can see it gradually doing something, like moving a ball across the
 * screen a little bit at a time. The <code>pause()</code> method will tell an 
 * <code>ActiveObject</code> to wait for some amount of time, allowing other 
 * <code>ActiveObject</code>s to work or allowing mouse and keyboard events to be processed.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class ActiveObject extends Thread {
	private static int id = 1;
	private boolean killed = false;
	
	private static final ThreadGroup GROUP = new ThreadGroup("ObjectDraw");
	
	/**
	 * Call the <code>kill()</code> method of every <code>ActiveObject</code>
	 * currently running. If your program is running as an applet in a web browser, 
	 * this method is automatically called when the user navigates away from the 
	 * page containing your applet. This prevents all your <code>ActiveObject</code>s
	 * from just merrily churning away with no applet left to draw on.
	 */
	public static void killAll() {
		final Thread[] threads = new Thread[GROUP.activeCount() + 20];
		final int length = GROUP.enumerate(threads);
		for(int i = 0; i < length; i++) {
			((ActiveObject)threads[i]).kill();
		}
	}
	
	/**
	 * This is the method to override to specify what work this <code>ActiveObject</code>
	 * will do when it is <code>start()</code>ed.
	 */
	public void run() {
		super.run();
	}
	
	/**
	 * Call this method to start this <code>ActiveObject</code> running. This method can only
	 * be called once for a particular object: even if it has finished its previous run, you
	 * cannot start it again. Instead, create and start a new instance.
	 */
	public void start() {
		super.start();
	}
	
	/**
	 * Kill this ActiveObject, stopping whatever it is doing. Because of the inherent danger
	 * in stopping a thread in the middle of running code, this method will merely set a
	 * flag saying this thread has been killed, which will then cause the thread to die the
	 * next time it calls <code>ActiveObject.pause()</code> or <code>ActiveObject.yield()</code>.
	 * <br><br>
	 * In order for this method to function properly, your <code>ActiveObject</code> ought
	 * to call <code>sleep()</code> or <code>yield()</code> appropriately rather than just
	 * running constantly. Also, if you catch <code>Throwable</code>, <code>Error</code>, 
	 * or <code>ThreadDeath</code> you will have to be sure to throw it again.
	 */
	public void kill() {
		killed = true;
		interrupt();
	}
	
	private static void checkKilled() {
		final Thread active = Thread.currentThread();
		if(active instanceof ActiveObject && ((ActiveObject)active).killed) throw new ThreadDeath();
	}
	
	/**
	 * If any other threads are waiting for a chance to do something, pause until they have
	 * had a  turn.
	 */
	public static void yield() {
		checkKilled();
		Thread.yield();
		checkKilled();
	}

	/**
	 * Pause for at least the number of milliseconds specified, plus the specified number of
	 * nanoseconds. A millisecond is a thousandth (1/1000) of a second, and a nanosecond is a
	 * billionth (1/1,000,000,000) of a second. So, <code>pause(1000, 0)</code> would pause for
	 * one second, and <code>pause(0, 500000)</code> would pause for half a millisecond.
	 * @param millis number of milliseconds to pause for
	 * @param nanos number of additional nanoseconds to pause for
	 */
	public static void pause(long millis, int nanos) {
		checkKilled();
		try {
			Thread.sleep(millis, nanos);
		} catch(InterruptedException e) {
		}
		checkKilled();
	}

	/**
	 * Pause for at least the number of milliseconds specified. A millisecond is a thousandth 
	 * (1/1000) of a second. So, <code>pause(1000, 0)</code> would pause for one second.
	 * @param millis number of milliseconds to pause for
	 */
	public static void pause(double millis) {
		pause((long)millis, 0);
	}

	/**
	 * Pause for at least the number of milliseconds specified. A millisecond is a thousandth 
	 * (1/1000) of a second. So, <code>pause(1000, 0)</code> would pause for one second.
	 * @param millis number of milliseconds to pause for
	 */
	public static void pause(long millis) {
		pause(millis, 0);
	}
	
	/**
	 * Create a new <code>ActiveObject</code>. The name of its thread in the debugger will be
	 * "ActiveObject" followed by a unique id number.
	 */
	public ActiveObject() {
		this("ActiveObject" + id++);
	}
	
	/**
	 * Create a new <code>ActiveObject</code> whose thread will show up in the debugger with
	 * the given name.
	 * @param name Name of thread to be shown in debugger
	 */
	public ActiveObject(String name) {
		super(GROUP, name);
	}
	
	/**
	 * Create a new <code>ActiveObject</code> that will run, not its own <code>run()</code>
	 * method, but the <code>run()</code> method of some other object that implements the
	 * <code>Runnable</code> interface.
	 * @param target An object whose <code>run()</code> method will be called when this
	 *               <code>ActiveObject</code> is <code>start()</code>ed, or <code>null</code>
	 *               if it should run its own <code>run()</code> method.
	 */
	public ActiveObject(Runnable target) {
		this(target, "ActiveObject" + id++);
	}
	
	/**
	 * Create a new <code>ActiveObject</code> that will run, not its own <code>run()</code>
	 * method, but the <code>run()</code> method of some other object that implements the
	 * <code>Runnable</code> interface.
	 * @param target An object whose <code>run()</code> method will be called when this
	 *               <code>ActiveObject</code> is <code>start()</code>ed, or <code>null</code>
	 *               if it should run its own <code>run()</code> method.
	 * @param name Thread name to represent this <code>ActiveObject</code> in the debugger.
	 */
	public ActiveObject(Runnable target, String name) {
		super(GROUP, target, name);
	}
}
