package objectdraw;

import java.awt.Color;

/**
 * The <code>AngLine</code> class provides a different way to make lines: instead of specifying
 * start and end points, you make an <code>AngLine</code> by providing a start point, the length
 * of the line, and the direction (as an angle in radians) that the line should extend from the
 * start point.
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class AngLine extends Line {
	/**
	 * Create a line of the specified <code>color</code> starting at <code>start</code> and 
	 * extending a distance <code>length</code> from there at an angle <code>radianAngle</code>.
	 * @param start <code>Location</code> of the start point of the line
	 * @param length Length of the line
	 * @param radianAngle Angular direction of the line, in radians. 0 is east, 
	 *                    <code>Math.PI / 2</code> is north, and so on, as is customary in math.
	 * @param color <code>Color</code> of the line
	 * @param canvas <code>DrawingCanvas</code> to add this line to
	 */
	public AngLine(Location start, double length, double radianAngle, Color color, DrawingCanvas canvas) {
		this(start.getDoubleX(), start.getDoubleY(), length, radianAngle, color, canvas);
	}
	
	/**
	 * Create a black line starting at <code>start</code> and 
	 * extending a distance <code>length</code> from there at an angle <code>radianAngle</code>.
	 * @param start <code>Location</code> of the start point of the line
	 * @param length Length of the line
	 * @param radianAngle Angular direction of the line, in radians. 0 is east, 
	 *                    <code>Math.PI / 2</code> is north, and so on, as is customary in math.
	 * @param canvas <code>DrawingCanvas</code> to add this line to
	 */
	public AngLine(Location start, double length, double radianAngle, DrawingCanvas canvas) {
		this(start, length, radianAngle, null, canvas);
	}

	/**
	 * Create a line of the specified <code>color</code> starting at (<code>x</code>, <code>y</code>) and 
	 * extending a distance <code>length</code> from there at an angle <code>radianAngle</code>.
	 * @param x x coordinate of the start location
	 * @param y y coordinate of the start location
	 * @param length Length of the line
	 * @param radianAngle Angular direction of the line, in radians. 0 is east, 
	 *                    <code>Math.PI / 2</code> is north, and so on, as is customary in math.
	 * @param color <code>Color</code> of the line
	 * @param canvas <code>DrawingCanvas</code> to add this line to
	 */
	public AngLine(double x, double y, double length, double radianAngle, Color color, DrawingCanvas canvas) {
		super(x, y, x + length * Math.cos(radianAngle), y + length * Math.sin(radianAngle), color, canvas);
	}
	
	/**
	 * Create a black line starting at (<code>x</code>, <code>y</code>) and 
	 * extending a distance <code>length</code> from there at an angle <code>radianAngle</code>.
	 * @param x x coordinate of the start location
	 * @param y y coordinate of the start location
	 * @param length Length of the line
	 * @param radianAngle Angular direction of the line, in radians. 0 is east, 
	 *                    <code>Math.PI / 2</code> is north, and so on, as is customary in math.
	 * @param canvas <code>DrawingCanvas</code> to add this line to
	 */
	public AngLine(double x, double y, double length, double radianAngle, DrawingCanvas canvas) {
		this(x, y, length, radianAngle, null, canvas);
	}
}
