package objectdraw;

import java.applet.AudioClip;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.RenderingHints;

import javax.swing.JApplet;
import javax.swing.JComponent;
import javax.swing.SwingUtilities;

/**
 * <code>Controller</code> is the class you should extend if you want to create an applet or
 * window that contains your own custom GUI instead of a <code>DrawingCanvas</code>. It provides
 * some helpful ObjectDraw features such as the <code>begin()</code> method, the 
 * <code>startController()</code> methods, and methods to load images and audio. Otherwise,
 * it leaves you free to create whatever GUI you want.
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Controller extends JApplet {
	static {
		try {
			System.setProperty("apple.laf.useScreenMenuBar", "true");
		} catch(Exception e) {
		}
	}
	private Dimension preferredSize;
	private GridPane grid;
	
	private boolean gridIsVisible = false;
    
    private static final Color MINOR_GRID_COLOR = new Color(0, 0, 0, 40);
    private static final Color MAJOR_GRID_COLOR = new Color(0, 0, 0, 80);
    
    // This component draws a grid over the controller's area of the frame, with 10 pixel grid spacing
    // and darker lines every 50.
    private class GridPane extends JComponent {
    	public GridPane() {
    		setOpaque(false);
    	}
    	
		public void paintComponent(Graphics graphics) {
			// The gird pane cannot be hidden, because then the rulers would not get
			// the mouse events they need to update the current cursor position. So,
			// we have to manually track if the grid is being drawn.
			if(!gridIsVisible) return;
			
			Graphics2D g = (Graphics2D)graphics;
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
			
			Point topLeft = SwingUtilities.convertPoint(Controller.this, new Point(0, 0), this);
			Point bottomRight = SwingUtilities.convertPoint(Controller.this, 
					new Point(Controller.this.getWidth(), Controller.this.getHeight()), this);
			
			g.setColor(MINOR_GRID_COLOR);
	        for(int i = 4, x = topLeft.x + 10; x < bottomRight.x; x += 10, i--) {
	        	if(i == 0) {
	        		g.setColor(MAJOR_GRID_COLOR);
	        		g.drawLine(x, topLeft.y, x, bottomRight.y);
	    			g.setColor(MINOR_GRID_COLOR);
	    			i = 5;
	        	} else {
	        		g.drawLine(x, topLeft.y, x, bottomRight.y);
	        	}
	        }
			
	        for(int i = 4, y = topLeft.y + 10; y < bottomRight.y; y += 10, i--) {
	        	if(i == 0) {
	        		g.setColor(MAJOR_GRID_COLOR);
	        		g.drawLine(topLeft.x, y, bottomRight.x, y);
	    			g.setColor(MINOR_GRID_COLOR);
	    			i = 5;
	        	} else {
	        		g.drawLine(topLeft.x, y, bottomRight.x, y);
	        	}
	        }
	        g.setColor(Color.BLACK);
		}
    }
    
    public Controller() {
    	super();
    }
    
    void makeGrid() {
    	if(grid != null) return;
    	
		grid = new GridPane();
		setGlassPane(grid);
		grid.setVisible(true);
    }
	
	/**
	 * Determine which version of ObjectDraw you are using.
	 * @return The ObjectDraw version, as a <code>String</code>.
	 */
	public static java.lang.String getVersion() {
		return "Unofficial - see www.zahniser.net/~russell/computer";
	}
	
	/**
	 * @deprecated internal use only
	 */
	public final void init() {
		finishLayout();
	}
	
	/**
	 * Override this method to do whatever your controller needs to do to set up its GUI.
	 */
	public void begin() {
		
	}
	
	/**
	 * @deprecated internal use only
	 */
	protected void finishLayout() {
		validate();
	    setVisible(true);
		repaint();
	}
	
	/**
	 * @deprecated internal use only
	 */
	protected void checkIfReady() {
		
	}
	
	/**
	 * @deprecated internal use only
	 */
	public void start() {
		checkIfReady();
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				begin();
			}
		});
	}
	
	/**
	 * @deprecated internal use only
	 */
	public void stop() {
		ActiveObject.killAll();
		EventTimer.stopAll();
	}
    
	/**
	 * @deprecated internal use only
	 */
    public Dimension getPreferredSize() {
    	return new Dimension(preferredSize);
    }
    
    private void setGridVisible(boolean visible) {
    	if(visible == gridIsVisible) return;
    	
    	gridIsVisible = visible;
    	if(grid != null) grid.repaint();
    }
    
    public void showGrid() {
    	setGridVisible(true);
    }
    
    public void hideGrid() {
    	setGridVisible(false);
    }
    
    public boolean isGridVisible() {
    	return gridIsVisible;
    }
	
	/**
	 * Run this <code>Controller</code> in a new window of the specified size. This method
	 * allows you to create and run a controller from a <code>main()</code> method, so that
	 * you program can run as an application rather than an applet. It also makes it possible
	 * to create and run a controller interactively in BlueJ or DrJava.
	 * @param width width of applet
	 * @param height height of applet
	 */
	public void startController(int width, int height) {
		startController(width, height, "ObjectDraw");
	}
	
	/**
	 * Run this <code>Controller</code> in a new window of the specified size. This method
	 * allows you to create and run a controller from a <code>main()</code> method, so that
	 * you program can run as an application rather than an applet. It also makes it possible
	 * to create and run a controller interactively in BlueJ or DrJava.
	 * @param width width of applet
	 * @param height height of applet
	 * @param name name of window
	 */
	public void startController(int width, int height, String name) {
		preferredSize = new Dimension(width, height);
		new ControllerFrame(name, this);
		init();
		start();
	}
	
	/**
	 * Run this <code>Controller</code> in a new window 400 pixels by 400 pixels. This method
	 * allows you to create and run a controller from a <code>main()</code> method, so that
	 * you program can run as an application rather than an applet. It also makes it possible
	 * to create and run a controller interactively in BlueJ or DrJava.
	 */
	public void startController() {
		startController(400, 400);
	}
	
	/**
	 * Load an audio file as an <code>AudioClip</code>. Not all file formats are guaranteed to
	 * work; in particular, the version of Java on your computer may not know how to play 
	 * <code>.mp3</code> files, even if you have a separate mp3 player. The formats that should
	 * almost always work are <code>.midi</code>, <code>.wav</code>, and <code>.au</code>.
	 * 
	 * @param path Path of the file, relative to the project directory. In other words, the sound
	 *             file should be inside your project folder.
	 * @return An <code>AudioClip</code> representing the requested file. If no such file was
	 *         found or it cannot be read, the return value will still be a valid <code>AudioClip</code>
	 *         object (not null), but it will not play anything.
	 */
	public AudioClip getAudio(String path) {
		return getAudioClip(getCodeBase(), path);
	}
	
	/**
	 * Load an image file as an <code>Image</code>. The image formats that should
	 * almost always work are <code>.png</code>, <code>.gif</code>, and <code>.jpg</code>.
	 * 
	 * @param path Path of the file, relative to the project directory. In other words, the image
	 *             file should be inside your project folder.
	 * @return An <code>Image</code> representing the requested file. If no such file was
	 *         found or it cannot be read, the return value will still be a valid <code>Image</code>
	 *         object (not null), but it will be invisible or may even cause an error when shown.
	 */
	public Image getImage(String path) {
		return getImage(getCodeBase(), path);
	}
}
