package objectdraw;

import java.awt.Component;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 * <code>MouseInterpreter</code> is the class that processes mouse events and calls the 
 * appropriate methods in <code>WindowController</code>. Additional mouse interpreters can be 
 * created to pipe these events to other objects as well. For example, you may want an 
 * ActiveObject to be controlled by the mouse, and you can do this by making a new 
 * <code>MouseInterpreter</code> that has that object as its target and attaching it to the 
 * canvas.
 * <br><br>
 * The methods that will be called by the interpreter are the following:
 * <br><br><code>
 * public void onMouseClick(Location point)<br>
 * public void onMouseDrag(Location point)<br>
 * public void onMouseMove(Location point)<br>
 * public void onMouseEnter(Location point)<br>
 * public void onMouseExit(Location point)<br>
 * public void onMousePress(Location point)<br>
 * public void onMouseRelease(Location point)
 * </code>
 * @author Russell Zahniser (russell@zahniser.net)
 *
 */
public class MouseInterpreter implements MouseListener, MouseMotionListener {
	private Component canvas;
	private Object target;
	
	/**
	 * Construct a <code>MouseInterpreter</code> that will listen for mouse events on the given
	 * <code>canvas</code> and relay them to the <code>target</code> object.
	 * @param target Object that will be notified of mouse events. It should have method such
	 * as <code>onMousePress()</code> for the events that it is interested in.
	 * @param canvas Canvas from which events will be captured.
	 */
	public MouseInterpreter(Object target, DrawingCanvas canvas) {
		this.canvas = (Component)canvas;
		this.canvas.addMouseListener(this);
		this.canvas.addMouseMotionListener(this);
		this.target = target;
	}
	
	/**
	 * Remove this <code>MouseInterpreter</code> from the canvas. It will no longer detect mouse 
	 * events, unless you add it back to a canvas.
	 */
	public void removeFromCanvas() {
		if(canvas != null) {
			canvas.removeMouseListener(this);
			canvas.removeMouseMotionListener(this);
			canvas = null;
		}
	}
	
	/**
	 * Add this <code>MouseInterpreter</code> to a canvas. It will automatically be removed from
	 * whatever canvas it was previously attached to.
	 * @param canvas <code>DrawingCanvas</code> to which this <code>MouseInterpreter</code>
	 * should listen.
	 */
	public void addToCanvas(DrawingCanvas canvas) {
		removeFromCanvas();
		this.canvas = (Component)canvas;
		this.canvas.addMouseListener(this);
		this.canvas.addMouseMotionListener(this);
	}
	
	/**
	 * Return the canvas this interpreter is attached to.
	 * @return the <code>DrawingCanvas</code> that this interpreter is attached to.
	 */
	public DrawingCanvas getCanvas() {
		return (DrawingCanvas)canvas;
	}

	private void mouseEvent(MouseEvent event, String type) {
		if(target == null) return;
		
		ObjectDrawObject.deferUpdates();
		Location point = new Location(event.getX(), event.getY());
		
		String methodName = "onMouse" + type;
		try {
			Method method = target.getClass().getMethod(methodName, new Class[] {Location.class});
			method.invoke(target, new Object[] {point});
		} catch(InvocationTargetException e) {
			e.getCause().printStackTrace();
		} catch (Exception e) {
		} finally {
			ObjectDrawObject.runUpdates();
		}
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void mouseDragged(MouseEvent event) {
		mouseEvent(event, "Drag");
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void mouseMoved(MouseEvent event) { 
		mouseEvent(event, "Move");
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void mouseClicked(MouseEvent event) { 
		mouseEvent(event, "Click");
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void mousePressed(MouseEvent event) { 
		canvas.requestFocus(); 
		mouseEvent(event, "Press");
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void mouseReleased(MouseEvent event) { 
		mouseEvent(event, "Release");
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void mouseEntered(MouseEvent event) { 
		mouseEvent(event, "Enter");
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void mouseExited(MouseEvent event) { 
		mouseEvent(event, "Exit");
	}
}
