package objectdraw;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.UnknownHostException;
import java.util.ArrayList;

import javax.swing.AbstractAction;
import javax.swing.AbstractListModel;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * The <code>Network</code> class will allow you to set up networked communication
 * between copies of your program running on different computers on a local area
 * network (LAN). You will always have to call <code>Network.configure()</code>
 * in order to specify how the client and server should run. You might use 
 * <code>startServer()</code> or <code>joinServer()</code> if you had a game
 * running on a web server. You should probably never use the other methods.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Network {
	static String gameIdentifier = null;
	static ServerListener serverListener = null;
	static ClientListener clientListener = null;
	static String lobbyAddress = "239.255.15.4";
	static int lobbyPort = 21504;
	static int gamePort = 21504;
	
	private static JDialog hostDialog = null;
	private static JDialog joinDialog = null;
	
	private static final ArrayList framesNeedingMenu = new ArrayList();
	private static final Action hostAction = new AbstractAction("Host network game") {
		public void actionPerformed(ActionEvent e) {
			openHostDialog();
		}
	};
	private static final Action joinAction = new AbstractAction("Join network game") {
		public void actionPerformed(ActionEvent e) {
			openJoinDialog();
		}
	};
	
	static void frameCreated(ControllerFrame frame) {
		if(isConfigured()) {
			addMenuToFrame(frame);
		} else {
			framesNeedingMenu.add(frame);
		}
	}
	
	private static void addMenuToFrame(ControllerFrame frame) {
		final JMenu menu = new JMenu("Network");
		
		menu.add(new JMenuItem(joinAction));
		menu.add(new JMenuItem(hostAction));
		
		frame.getJMenuBar().add(menu);
	}
	
	/**
	 * Configure the parameters that will make it possible for this program to play
	 * against other copies of itself over a network. Once this method has been called,
	 * the Network menu will show up in your ObjectDraw windows.
	 * <br><br>
	 * As a general rule, this method should be called once, and <b>only</b> once, 
	 * as soon as possible after your program starts (anywhere in the begin() method 
	 * is fine). If you reconfigure the server in between creating a server and a 
	 * client, they won't be able to talk to each other properly, so setting the 
	 * configuration once and then not messing with it again is the best policy.
	 * <br><br>
	 * Please note that networking is not possible with applets, only Java applications 
	 * - that is, programs that have a main() method.
	 * 
	 * @param game Unique name for this game that identifies this particular program.
	 *        This will prevent someone from accidentally connecting to a different
	 *        game that doesn't know how to respond.
	 * @param server Object that will deal with the server side of the program. If you
	 *        just want the server to take each message that it gets and repeat it back
	 *        to all the clients, you can use <code>ServerListener.ECHO</code>. Otherwise,
	 *        you should have some class of your own implement <code>ServerListener</code>.
	 * @param client Object that will deal with the client side of the program. Often this
	 *        will be your <code>WindowController</code>; you can just declare it to
	 *        implement <code>ClientListener</code> and then have the client methods
	 *        draw things on the canvas.
	 */
	public static void configure(final String game, final ServerListener server, final ClientListener client) {
		if(game == null) throw new IllegalArgumentException("The game identifier cannot be null.");
		if(server == null) throw new IllegalArgumentException("The server listener cannot be null.");
		if(client == null) throw new IllegalArgumentException("The client listener cannot be null.");
		
		gameIdentifier = game;
		serverListener = server;
		clientListener = client;
		
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				for(int i = 0; i < framesNeedingMenu.size(); i++) {
					addMenuToFrame((ControllerFrame)framesNeedingMenu.get(i));
				}
				framesNeedingMenu.clear();
			}
		});
	}
	
	private static boolean isConfigured() {
		return gameIdentifier != null;
	}
	
	static void closeServer(final Server server) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() { hostAction.setEnabled(true); }
		});
	}
	
	static void closeClient(final Client client) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() { joinAction.setEnabled(true); }
		});
	}
	
	/**
	 * Have this program start up a game server. The <code>name</code> given will show up 
	 * in the list under the "Join network game" menu option. This method may only be called
	 * after you have done <code>Network.configure()</code>.
	 * @param name Server name
	 */
	public static void startServer(final String name) {
		if(!isConfigured()) throw new IllegalStateException("You must call configure() first.");
		if(!hostAction.isEnabled()) throw new IllegalStateException("You may only have one server running at a time.");
		
		startServer(name, null);
	}
	
	private static void startServer(final String name, final Runnable onLoad) {
		try {
			new Server(name, onLoad);
			hostAction.setEnabled(false);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private static void connectToServer(final InetAddress address, final String name) {
		if(!joinAction.isEnabled()) throw new IllegalStateException("You may only have one server running at a time.");
		
		try {
			new Client(name, address);
			joinAction.setEnabled(false);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Connect to a server at a particular IP address. This might be used if you have
	 * a server running on a web server at a known address on the internet.
	 * @param address IP address (in the form "XX.XX.XX.XX") or machine name (in 
	 *        the form "www.zahniser.net") of server
	 * @param name Client name
	 */
	public static void connectToServer(final String address, final String name) {
		if(!isConfigured()) throw new IllegalStateException("You must call configure() first.");
		
		try {
			connectToServer(InetAddress.getByName(address), name);
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
	}
	
	private static class HostListener extends WindowAdapter implements CaretListener, ActionListener, ItemListener {
		private final JDialog dialog;
		private final JButton cancel, host;
		private final JCheckBox joinBox;
		private final JTextField gameName, joinName;
		
		public HostListener(final JDialog dialog, final JButton cancel, final JButton host,
				final JCheckBox joinBox, final JTextField gameName, final JTextField joinName) {
			this.dialog = dialog;
			this.cancel = cancel;
			this.host = host;
			this.joinBox = joinBox;
			this.gameName = gameName;
			this.joinName = joinName;
			
			dialog.addWindowListener(this);
			cancel.addActionListener(this);
			host.addActionListener(this);
			joinBox.addItemListener(this);
			gameName.addCaretListener(this);
			joinName.addCaretListener(this);
			
			checkCanJoin();
		}
		
		private void checkCanJoin() {
			host.setEnabled(gameName.getText().length() > 0 && 
					(!joinBox.isSelected() || joinName.getText().length() > 0));
		}
		
		public void caretUpdate(final CaretEvent e) {
			checkCanJoin();
		}

		public void windowActivated(final WindowEvent e) {
			gameName.requestFocusInWindow();
		}
		
		public void windowDeactivated(final WindowEvent e) {
			gameName.setText("");
			joinName.setText("");
			joinBox.setSelected(true);
		}

		public void actionPerformed(final ActionEvent e) {
			if(e.getSource() == host) {
				final String joinAs = joinName.getText();
				final Runnable onLoad = (!joinBox.isSelected() ? null : new Runnable() {
					public void run() {
						connectToServer("127.0.0.1", joinAs);
					}
				});
				startServer(gameName.getText(), onLoad);
			}
			dialog.hide();
		}

		public void itemStateChanged(ItemEvent e) {
			if(e.getStateChange() == ItemEvent.SELECTED) {
				joinName.setEnabled(true);
			} else {
				joinName.setText("");
				joinName.setEnabled(false);
			}
			checkCanJoin();
		}
	}
	
	private static JDialog createHostDialog() {
		final JDialog dialog = new JDialog();
		dialog.setTitle((String)hostAction.getValue(Action.NAME));
		dialog.setModal(true);
		dialog.setDefaultCloseOperation(JDialog.HIDE_ON_CLOSE);
		
		final JPanel content = new JPanel();
		dialog.setContentPane(content);
		content.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		content.setLayout(new BoxLayout(content, BoxLayout.Y_AXIS));
		
		final JPanel namePanel = new JPanel();
		namePanel.setLayout(new BoxLayout(namePanel, BoxLayout.X_AXIS));
		namePanel.add(new JLabel("Game name: "));
		final JTextField gameName = new JTextField();
		namePanel.add(gameName);
		content.add(namePanel);
		
		content.add(Box.createVerticalStrut(10));
		
		final JPanel joinPanel = new JPanel();
		joinPanel.setLayout(new BoxLayout(joinPanel, BoxLayout.X_AXIS));
		final JCheckBox joinBox = new JCheckBox("Also join as: ");
		joinBox.setSelected(true);
		joinPanel.add(joinBox);
		final JTextField joinName = new JTextField();
		joinPanel.add(joinName);
		content.add(joinPanel);
		
		content.add(Box.createVerticalStrut(20));
		
		final JPanel buttonPanel = new JPanel();
		buttonPanel.setLayout(new BoxLayout(buttonPanel, BoxLayout.X_AXIS));
		buttonPanel.add(Box.createHorizontalStrut(150));
		final JButton cancel = new JButton("Cancel");
		buttonPanel.add(cancel);
		final JButton host = new JButton("Host");
		dialog.getRootPane().setDefaultButton(host);
		buttonPanel.add(host);
		content.add(buttonPanel);
		
		new HostListener(dialog, cancel, host, joinBox, gameName, joinName);
		
		dialog.pack();
		return dialog;
	}
	
	/**
	 * This is a way for your program to automatically open the "Host network game" dialog,
	 * for those who are too cool to use the menus.
	 */
	public static void openHostDialog() {
		if(hostDialog == null) hostDialog = createHostDialog();
		
		hostDialog.show();
	}
	
	
	private static class DiscoveryThread extends AbstractListModel implements Runnable {
		private static final int DATA_LENGTH = 256;
		private static final int SERVER_TIMEOUT = 1000;
		
		private final DatagramPacket dgram;
		private final MulticastSocket socket;
		private final ArrayList discoveries;
		private boolean cancelled = false;
		
		public class Discovery implements Runnable {
			private final String serverName;
			private final InetAddress serverAddress;
			private long timeStamp;
			
			public Discovery(final String serverName, final InetAddress serverAddress) {
				this.serverName = serverName;
				this.serverAddress = serverAddress;
				
				timeStamp = System.currentTimeMillis();
				
				SwingUtilities.invokeLater(this);
			}
			
			public void run() {
				checkDiscovery(this);
			}
			
			public String toString() {
				return serverName;
			}
		}
		
		public DiscoveryThread() throws IOException {
			discoveries = new ArrayList();
			
			dgram = new DatagramPacket(new byte[DATA_LENGTH], DATA_LENGTH);
			socket = new MulticastSocket(lobbyPort);
			socket.joinGroup(InetAddress.getByName(lobbyAddress));
		}

		public void run() {
			try {
				while(!cancelled) {
					socket.receive(dgram);
					final String message = new String(dgram.getData(), dgram.getOffset(), dgram.getLength(), NetworkThread.CHARSET_NAME);
					
					final String serverName = message.substring(0, message.indexOf('\n'));
					final String gameName = message.substring(message.lastIndexOf('\n') + 1);
					if(gameName.equals(gameIdentifier)) new Discovery(serverName, dgram.getAddress());
					
					dgram.setLength(DATA_LENGTH);
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
			cancelled = false;
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					clear();
				}
			});
		}
		
		// This will always be called on the event-dispatch thread, so we are free to modify the model
		// First, check if any discoveries are out-of-date, and remove them; then, add the new one if
		// you have not seen a duplicate. If you do see a duplicate, refresh its timestamp.
		public void checkDiscovery(final Discovery discovery) {
			// -2 will stand for "at the end"
			int insertIndex = -2;
			
			for(int i = 0; i < discoveries.size(); i++) {
				final Discovery old = (Discovery)discoveries.get(i);
				
				if(old.serverAddress.equals(discovery.serverAddress)) {
					old.timeStamp = discovery.timeStamp;
					// -1 will stand for "Don't have to insert"
					insertIndex = -1;
				} else if(old.timeStamp + SERVER_TIMEOUT < discovery.timeStamp) {
					// This server is no longer announcing - remove it
					discoveries.remove(i);
					fireIntervalRemoved(this, i, i);
					i--;
				} else if(old.serverName.compareTo(discovery.serverName) > 0) {
					// insert the new discovery before this one
					insertIndex = i;
				}
			}
			
			if(insertIndex == -2) insertIndex = discoveries.size();
			if(insertIndex >= 0) {
				discoveries.add(insertIndex, discovery);
				fireIntervalAdded(this, insertIndex, insertIndex);
			}
		}
		
		public void cancel() {
			cancelled = true;
		}
		
		public void clear() {
			final int length = discoveries.size();
			discoveries.clear();
			fireIntervalRemoved(this, 0, length - 1);
		}

		public Object getElementAt(int index) {
			return discoveries.get(index);
		}

		public int getSize() {
			return discoveries.size();
		}
	}
	
	private static class JoinListener extends WindowAdapter implements ListSelectionListener, CaretListener, ActionListener, ListDataListener {
		private final DiscoveryThread discoverer;
		private final JDialog dialog;
		private final JButton join;
		private final JButton cancel;
		private final JList list;
		private final JTextField joinName;
		
		public JoinListener(final JDialog dialog, final JList list, 
				final JTextField joinName, final JButton cancel, final JButton join) throws IOException {
			
			this.discoverer = new DiscoveryThread();
			this.dialog = dialog;
			this.join = join;
			this.cancel = cancel;
			this.list = list;
			this.joinName = joinName;
			
			list.setModel(discoverer);
			discoverer.addListDataListener(this);
			dialog.addWindowListener(this);
			list.addListSelectionListener(this);
			joinName.addCaretListener(this);
			cancel.addActionListener(this);
			join.addActionListener(this);
			
			checkCanJoin();
		}
		
		public void actionPerformed(final ActionEvent e) {
			if(e.getSource() == join) {
				DiscoveryThread.Discovery server = (DiscoveryThread.Discovery)list.getSelectedValue();
				if(server != null) {
					connectToServer(server.serverAddress, joinName.getText());
				}
			}
			dialog.hide();
		}

		public void windowActivated(final WindowEvent e) {
			new Thread(discoverer).start();
			joinName.requestFocusInWindow();
		}
		
		public void windowDeactivated(final WindowEvent e) {
			joinName.setText("");
			discoverer.cancel();
		}
		
		private void checkCanJoin() {
			join.setEnabled(list.getSelectedValue() != null && joinName.getText().length() > 0);
		}

		public void valueChanged(ListSelectionEvent e) {
			checkCanJoin();
		}

		public void caretUpdate(CaretEvent e) {
			checkCanJoin();
		}

		public void contentsChanged(ListDataEvent e) { }
		public void intervalRemoved(ListDataEvent e) { }

		public void intervalAdded(ListDataEvent e) {
			if(list.getSelectedValue() == null) {
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						list.setSelectedIndex(0);
					}
				});
			}
		}
	}
	
	private static JDialog createJoinDialog() {
		final JDialog dialog = new JDialog();
		dialog.setTitle((String)hostAction.getValue(Action.NAME));
		dialog.setModal(true);
		dialog.setDefaultCloseOperation(JDialog.HIDE_ON_CLOSE);
		
		final JPanel content = new JPanel();
		dialog.setContentPane(content);
		content.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		content.setLayout(new BoxLayout(content, BoxLayout.Y_AXIS));
		
		final JList hostList = new JList();
		content.add(new JScrollPane(hostList));
		
		content.add(Box.createVerticalStrut(10));
		
		final JPanel namePanel = new JPanel();
		namePanel.setLayout(new BoxLayout(namePanel, BoxLayout.X_AXIS));
		namePanel.add(new JLabel("Join as: "));
		final JTextField joinName = new JTextField();
		namePanel.add(joinName);
		content.add(namePanel);
		
		content.add(Box.createVerticalStrut(20));
		
		final JPanel buttonPanel = new JPanel();
		buttonPanel.setLayout(new BoxLayout(buttonPanel, BoxLayout.X_AXIS));
		buttonPanel.add(Box.createHorizontalStrut(150));
		final JButton cancel = new JButton("Cancel");
		buttonPanel.add(cancel);
		final JButton join = new JButton("Join");
		dialog.getRootPane().setDefaultButton(join);
		buttonPanel.add(join);
		content.add(buttonPanel);
		
		try {
			new JoinListener(dialog, hostList, joinName, cancel, join);
		} catch(IOException e) {
			e.printStackTrace();
		}
		
		dialog.pack();
		return dialog;
	}
	
	/**
	 * This is a way for your program to automatically open the "Join network game" dialog,
	 * for those who are too cool to use the menus.
	 */
	public static void openJoinDialog() {
		if(joinDialog == null) joinDialog = createJoinDialog();
		
		joinDialog.show();
	}
	
	/**
	 * Set the IP address of the UDP multicast group used to discover active games when
	 * someone wants to join a game. You will hopefully never need to mess with this.
	 * If you are having trouble getting your program to connect over your LAN, messing
	 * with this is not the solution - configuring your computer's firewall is. The default
	 * address is 239.255.15.4.
	 * @param address Class D IP address for the multicast group
	 */
	public static void setLobbyAddress(final String address) {
		lobbyAddress = address;
	}
	
	/**
	 * Set which port is used for the UPD multicast that identifies active games for
	 * someone to join. If you need to change this, it probably means you are on a 
	 * system with a restrictive firewall in place, and it ought to also mean
	 * that you have some clue what you're doing. The default port is 21504.
	 * @param port 
	 */
	public static void setLobbyPort(final int port) {
		lobbyPort = port;
	}
	
	/**
	 * Set which port the game should use. If you need to change this, it probably means
	 * you are on a system with a restrictive firewall in place, and it ought to also mean
	 * that you have some clue what you're doing. The default port is 21504.
	 * @param port
	 */
	public static void setGamePort(final int port) {
		gamePort = port;
	}
}
