package objectdraw;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.DatagramChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.ServerSocketChannel;

import javax.swing.SwingUtilities;
import javax.swing.Timer;

/**
 * A <code>Server</code> runs on one computer and allows other computers (clients)
 * to connect to it. The clients can then send messages to the server and receive 
 * messages back. The server also announces its presence across the local area
 * network (LAN) so that computers at the same site can see that a game is happening.
 * <br><br>
 * You will never create a <code>Server</code> object directly. Instead, write a 
 * class that implements <code>ServerListener</code> and pass an object of that
 * class to <code>Network.configure()</code>. This will result in that object 
 * being informed when server-related events happen.
 * <br><br>
 * The <code>Server</code> class contains methods to send messages either to a
 * particular client, or to all the clients at once. Your <code>ServerListener</code>
 * will be passed a <code>Server</code> object with each event; you could also keep
 * it in an instance variable so that you can send messages whenever you want.
 * 
 * @see ServerListener
 * @see Network#configure(String, ServerListener, ClientListener)
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class Server extends NetworkThread {
	private final ServerSocketChannel serverSocket;
	private final ServerListener listener;
	private final MulticastAnnouncerChannel announcer;
	private final Runnable onLoad;
	
	Server(final String name, final Runnable onLoad) throws IOException {
		super(name);
		
		this.onLoad = onLoad;
		
		serverSocket = ServerSocketChannel.open();
		announcer = new MulticastAnnouncerChannel();
	    
	    listener = Network.serverListener;
	    
	    start();
	}
	
	/**
	 * Send a message to a single client. This will result in the client listener 
	 * on that computer getting a <code>messageReceived()</code> event. The channel
	 * ID is the number identifying that client that is passed to the listener along
	 * with client-specific events.
	 * @param message The message to send.
	 * @param channelID The ID of the client to send the message to.
	 */
	public void sendMessage(final String message, final int channelID) {
		enqueueMessage(message, channelID);
	}
	
	/**
	 * Send a message to every client connected to the server. This will result 
	 * in the client listener getting a <code>messageReceived()</code> event on
	 * all the computers connected to this server.
	 * @param message The message to send.
	 */
	public void sendMessage(final String message) {
		enqueueMessage(message, -1);
	}
	
	/**
	 * Get the server name. This is the name that the user entered in the
	 * "Host a network game" window.
	 * @return the server name
	 */
	public String getName() {
		return super.getName();
	}
	
	/**
	 * Check if the server is still running.
	 * @return <code>true</code> if the server is running, <code>false</code> otherwise.
	 */
	public boolean isRunning() {
		return super.isRunning();
	}
	
	/**
	 * Shut down the server, disconnecting all clients. Each client listener
	 * will receive a <code>clientDisconnected()</code> message, and the server
	 * listener will receive a <code>channelClosed()</code> event for every client,
	 * followed by a <code>serverStopped()</code> event. Quitting the program will
	 * automatically shut down the server.
	 */
	public void shutDown() {
		super.shutDown();
	}
	
	void messageReceived(final int channel, final String channelName, final String message) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				listener.messageReceived(Server.this, channel, channelName, message);
			}
		});
	}
	
	void channelClosed(final int channel, final String channelName) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				listener.channelClosed(Server.this, channel, channelName);
			}
		});
	}
	
	void channelReady(final int channel, final String channelName) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				listener.channelOpened(Server.this, channel, channelName);
			}
		});
	}
	
	private class MulticastAnnouncerChannel implements ChannelHandler, ActionListener {
		private final DatagramChannel channel;
		private final Timer timer;
		private final ByteBuffer message;
		
		// This channel just announces; it never reads.
		public MulticastAnnouncerChannel() throws IOException {
			channel = DatagramChannel.open();
			timer = new Timer(100, this);
			message = ByteBuffer.wrap((getName() + "\n" + Network.gameIdentifier).getBytes(CHARSET_NAME));
		}
		
		public void setup() throws IOException  {
			channel.configureBlocking(false);
			channel.connect(new InetSocketAddress(Network.lobbyAddress, Network.lobbyPort));
			channel.register(selector, 0, this);      
	        
	        timer.start();
		}
		
		public void close() {
			timer.stop();
			try {
				channel.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		public void read() { }
		
		public void write() {
			try {
				channel.write(message);
			} catch (IOException e) {
				e.printStackTrace();
				close();
			}
			
			if(message.remaining() == 0) {
				message.rewind();
				updateChannelLater(channel, 0, SelectionKey.OP_WRITE);
			}
		}

		public void actionPerformed(ActionEvent e) {
			updateChannelLater(channel, SelectionKey.OP_WRITE, SelectionKey.OP_WRITE);
			selector.wakeup();
		}
	}

	boolean setup() {
		// This initial configuration, plus the way pending messages are handled,
		// is essentially all that is different between a client and a server.
		// We should extract the shared functionality into a superclass.
		if(!super.setup()) return false;

        try {
			serverSocket.configureBlocking(false);
		    final InetSocketAddress isa = new InetSocketAddress(Network.gamePort);
		    serverSocket.socket().bind(isa, 100);
		    
	        serverSocket.register(selector, SelectionKey.OP_ACCEPT, null);      
		} catch (IOException e) {
			e.printStackTrace();
			closeAll();
			return false;
		}
		
		try {
			announcer.setup();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				listener.serverStarted(Server.this);
				if(onLoad != null) onLoad.run();
			}
		});
		
		return true;
	}
	
	void tearDown() {
		try {
			serverSocket.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				listener.serverStopped(Server.this);
			}
		});
		
		Network.closeServer(this);
		
		super.tearDown();
	}
}
