package objectdraw;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.font.FontRenderContext;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.text.AttributedString;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * <code>TextWrappingLayout</code> is a class that can perform text layout wrapping to any arbitrary shape, and
 * aligning any arbitrary amount in both the x and the y direction. Layout proceeds in left-to-right
 * rows from the top to the bottom of the shape, with the first row positioned wherever there is space
 * to fit the first word of the text, and each row possibly consisting of several runs in different
 * parts of the shape if the shape has several discrete interior regions along that row. A line break
 * will be inserted at any newline ("\n") character in the text. Text can consist of many runs, each
 * with its own font and color; each run will be given a shape to draw once layout is complete.
 * <br><br>
 * If all you want is to draw text with certain styles and colors wrapped in a shape, consider using the
 * <code>SimpleClassRun</code> inner class. If you already have your own class that stores information about the
 * text to be displayed, or if you want to be able to do something with the <code>Shape</code> produced by the layout
 * other than simply drawing it, consider making it implement <code>TextRun</code> instead; this gives you more flexibility.
 * <br><br>
 * If you do not need your text to be bounded by a shape, you can simply pass null as the shape. The text
 * will then be laid out aligned properly around (0, 0), and it will break only where you explicitly put
 * a newline character in your text.
 * @deprecated for internal use only
 * @author Russell Zahniser (russell@zahniser.net)
 * @version 20070629
 */
public class TextWrappingLayout {
	private TextRun start;
	private double alignV;
	private double alignH;
	private Shape shape;
	private FontRenderContext context;
	private double endPadding = 4;
	
	private static final double MIN_ERROR = .5;
	private static final int WORKING = 0;
	private static final int GROW = 1;
	private static final int TEXT_END = 2;
	private static final int LINE_END = 3;
	
	public static final double ALIGN_TOP = -1;
	public static final double ALIGN_BOTTOM = 1;
	public static final double ALIGN_LEFT = -1;
	public static final double ALIGN_RIGHT = 1;
	public static final double ALIGN_CENTER = 0;
	
	/**
	 * A TextRun is a segment of text with a single font and color. TextRun is meant to be implemented
	 * as a linked list, with each run having a link to the next. Once layout has successfully completed, 
	 * each run will have its shape set; it can then draw itself by simply setting its own color and 
	 * drawing that shape. Text will automatically be broken at newlines ("\n").
	 * @deprecated for internal use only
	 */
	public static interface TextRun {
		public void setShape(Shape shape);
		public Font getFont();
		public String getText();
		public TextRun getNextRun();
	}
	
	/**
	 * SimpleTextRun is a basic implementation of TextRun that can be used
	 * to simply draw styled text to a graphics wrapped around a shape.
	 * @deprecated for internal use only
	 */
	public static class SimpleTextRun implements TextRun {
		private Font font;
		private Color color;
		private String text;
		private SimpleTextRun next;
		private Shape shape;
		
		/**
		 * Create a single run of text in a single font and color. You can create several
		 * runs this way and append them one by one to the first run to get a long text
		 * with several style runs.
		 * @param text Text that will be drawn in this font and color
		 * @param font Font that this text will be displayed in
		 * @param color Color that this text will be drawn in
		 */
		public SimpleTextRun(String text, Font font, Color color) {
			this.text = text;
			this.font = font;
			this.color = color;
		}
		
		/**
		 * @see TextWrappingLayout.TextRun#getFont() TextRun.getFont()
		 */
		public Font getFont() {
			return font;
		}
		
		/**
		 * @see TextWrappingLayout.TextRun#getNextRun() TextRun.getNextRun()
		 */
		public TextRun getNextRun() {
			return next;
		}

		/**
		 * @see TextWrappingLayout.TextRun#getText() TextRun.getText()
		 */
		public String getText() {
			return text;
		}

		/**
		 * @see TextWrappingLayout.TextRun#setShape(java.awt.Shape) TextRun.setShape()
		 */
		public void setShape(Shape shape) {
			this.shape = shape;
		}
		
		/**
		 * Add another run of text at the end.
		 * @param run Run to add
		 */
		public void append(SimpleTextRun run) {
			if(next == null) {
				next = run;
			} else {
				next.append(run);
			}
		}
		
		/**
		 * Wrap this text into the given shape, with the given vertical and horizontal alignment.
		 * This must be called first as a preparation for calling draw()
		 * @param shape Shape to wrap this text into.
		 * @param alignH Horizontal alignment, as a double. -1 aligns along left, 1 aligns along right, and
		 *        anything in between skews alignment that much one way or the other. 0 is centered.
		 * @param alignV Vertical alignment, as a double. -1 aligns along top, 1 aligns along bottom, and
		 *        anything in between skews alignment that much one way or the other. 0 is centered.
		 */
		public void prepareLayout(Shape shape, double alignH, double alignV, FontRenderContext frc) {
			new TextWrappingLayout(this, alignH, alignV, shape, frc).doLayout();
		}
		
		/**
		 * Draw this text in the given graphics. You must have first called prapareLayout() to
		 * lay out the text.
		 * @param g Graphics2D to draw into. (The Graphics passed as a parameter to the paint() methods
		 *        of any Swing class is actually a Graphics2D; you can just typecast it)
		 */
		public void draw(Graphics2D g) {
			if(shape != null) {
				g.setColor(color);
				g.fill(shape);
			}
			if(next != null) {
				next.draw(g);
			}
		}
	}
	
	/**
	 * Create a TextWrappingLayout that will take its text from the given TextRun and produce a layout 
	 * in the given shape, with the given alignment. No data is cached from the TextRun, so if it wants
	 * to change its text or font and try calling doLayout() again, this is possible. This allows, for
	 * example, text altering its size until it fits in the layout space.
	 * @param start TextRun that begins the linked list of text to be written
	 * @param alignH Horizontal alignment, as a double. -1 aligns along left, 1 aligns along right, and
	 *        anything in between skews alignment that much one way or the other. 0 is centered.
	 * @param alignV Vertical alignment, as a double. -1 aligns along top, 1 aligns along bottom, and
	 *        anything in between skews alignment that much one way or the other. 0 is centered.
	 * @param shape Shape to wrap the given text run into
	 */
	public TextWrappingLayout(TextRun start, double alignH, double alignV, Shape shape, FontRenderContext frc) {
		this.start = start;
		this.alignV = alignV;
		this.alignH = alignH;
		this.shape = shape;
		this.context = frc;
	}
	
	/**
	 * Text is automatically inset somewhat from the edges of each row. The default amout is just four
	 * pixels at each end. Use this method to set it to a larger or smaller amount of padding.
	 * @param padding Amount to inset text by at either end
	 */
	public void setEndPadding(double padding) {
		endPadding = padding;
	}
	
	// Right now this doesn't actually work - someday I'll fill it in
	private boolean doLayoutOnArc() {
		// Construct an AttributedString of the entire text
		TextRun run = start;
		String text = "";
		while(run != null) {
			text += run.getText();
			run = run.getNextRun();
		}
		if(text.length() == 0) {
			writeLayout(null);
			return true;
		}
		AttributedString string = new AttributedString(text);
		int offset = 0;
		run = start;
		while(run != null) {
			string.addAttribute(TextAttribute.FONT, run.getFont(), offset, offset + run.getText().length());
			offset += run.getText().length();
			run = run.getNextRun();
		}
		
		String[] lines = text.split("\n");
		int end = text.length(), start = end;
		double height = 0;
		TextLayout layout = null;
		for(int j = lines.length - 1; j >= 0; j--) {
			String line = lines[j];
			end = start;
			start -= line.length();
			
			while(start < end && Character.isWhitespace(text.charAt(start))) start++;
			while(start < end && Character.isWhitespace(text.charAt(end))) end--;
			
			if(start == end) continue;
			
			layout = new TextLayout(string.getIterator(null, start, end), context);
			height += layout.getAscent() + layout.getDescent() + layout.getLeading();
			
			// Using the length of the baseline of this layout, we will do the actual laying out one
			// character at a time, with the baseline going along the ellipse and the length of the
			// advance corresponding to the chord length covered by each character.
		}
		height -= layout.getLeading();
		writeLayout(null);
		return true;
	}
	
	// Layout without a shape to wrap to - simply keep the origin at the correctly aligned position
	// and break only at newlines.
	private boolean doLayoutNoShape() {
		// Construct an AttributedString of the entire text
		TextRun run = start;
		String text = "";
		while(run != null) {
			text += run.getText();
			run = run.getNextRun();
		}
		if(text.length() == 0) {
			writeLayout(null);
			return true;
		}
		AttributedString string = new AttributedString(text);
		int offset = 0;
		run = start;
		while(run != null) {
			int len = run.getText().length();
			if(len > 0) {
				string.addAttribute(TextAttribute.FONT, run.getFont(), offset, offset + len);
				offset += len;
			}
			run = run.getNextRun();
		}
		
		// Go through one line at a time, but make separate paths for each run
		GeneralPath path = new GeneralPath();
		ArrayList paths = new ArrayList();
		run = start;
		int runEndOffset = run.getText().length();
		while(runEndOffset == 0) {
			run = run.getNextRun();
			runEndOffset = run.getText().length();
		}
		offset = 0; // Current position in line
		double top = 0;
		double spacing = 0; // Spacing to add on newline
		TextLayout layout = null;
		while(run != null) {
			// Trim whitespace at the start of the line, making a newline for every \n
			char c;
			while(offset < text.length() && Character.isWhitespace(c = text.charAt(offset))) {
				if(c == '\n') top += spacing;
				offset++;
			}
			// If we found whitespace up to the end of the text, quit now.
			if(offset == text.length()) {
				paths.add(path);
				break;
			}
			
			// Find out where the end of this line is
			int endOffset = text.indexOf("\n", offset);
			if(endOffset == -1) {
				endOffset = text.length();
			} else {
				endOffset++;
			}
			
			// Layout for whole line allows us to figure out how to position it
			layout = new TextLayout(string.getIterator(null, offset, endOffset), context);
			
			AffineTransform transform = AffineTransform.getTranslateInstance(
					-layout.getBounds().getMinX() - (1 + alignH) * layout.getBounds().getWidth() / 2, 
					top + layout.getAscent());
			// Create a separate shape for each run
			while(offset < endOffset) {
				// layout either to end of run or end of line
				int end = Math.min(runEndOffset, endOffset);
				if(offset != end) {
					TextLayout runLayout = new TextLayout(string.getIterator(null, offset, end), context);
					
					path.append(runLayout.getOutline(transform), false);
					transform.translate(runLayout.getAdvance(), 0);
					offset = end;
				}
				
				if(offset == runEndOffset) {
					// I got to the end of a run
					run = run.getNextRun();
					paths.add(path);
					if(run != null) {
						runEndOffset += run.getText().length();
						path = new GeneralPath();
					}
				}
			}
			top += layout.getAscent() + layout.getDescent() + layout.getLeading();
			spacing = ((layout.getAscent() + layout.getDescent()) / 2 + layout.getLeading());
		}
		top -= layout.getLeading();
		
		// As we write the shapes to the runs, offset them appropriately in the vertical direction.
		AffineTransform transform = AffineTransform.getTranslateInstance(0, -(1 + alignV) * top / 2);
		for(run = start, offset = 0; run != null; run = run.getNextRun(), offset++) {
			run.setShape(transform.createTransformedShape((Shape)paths.get(offset)));
		}
		return true;
	}
	
	private boolean layoutInRect(Rectangle2D rect) {
		HashMap map = tryLayout(rect.getMinY() + endPadding);
		double top = getMapTop(map), bottom = getMapBottom(map), height = bottom - top;
		double wiggleRoom = rect.getHeight() - 2 * endPadding - height;
		
		AffineTransform transform = AffineTransform.getTranslateInstance(0, (1 + alignV) * wiggleRoom / 2);
		writeLayout(map, transform);
		return isFullLayout(map);
	}
	
	/**
	 * Attempt to layout your text in your shape. Return true or false indicating whether the entire
	 * text was successfully fit into the shape.
	 * @return true if all text fit, false if it was broken off at the end
	 */
	public boolean doLayout() {
		WordInfo info = getFirstWord(start, 0);
		if(!info.exists) {
			writeLayout(null);
			return true;
		}
		
		if(shape == null) return doLayoutNoShape();
		else if((shape instanceof Arc2D) && ((Arc2D)shape).getArcType() == Arc2D.OPEN) return doLayoutOnArc();
		else if(shape instanceof Rectangle2D) return layoutInRect((Rectangle2D)shape);
		
		Rectangle2D firstLine = placeWord(info.width, info.height, -1);
		if(firstLine == null) {
			writeLayout(null);
			return false;
		}
		
		double minY = firstLine.getMinY();
		HashMap map = tryLayout(minY);
		if(alignV == ALIGN_TOP || !isFullLayout(map)) {
			writeLayout(map);
			return isFullLayout(map);
		}
		
		double top = shape.getBounds2D().getMinY();
		double bottom = shape.getBounds2D().getMaxY();
		
		// Search for the alignment where the center best matches alignV.
		double maxY = placeWord(info.width, info.height, 1).getMinY();
		double align = (1 + alignV) / 2;
		HashMap bestMap = map;
		double bestAlign = (getMapTop(map) - top) / (getMapTop(map) - top + bottom - getMapBottom(map));
		// If we find a new align that is better than this, we can start narrowing the search.
		// The procedure will be: if the new best align is greater than align, we can use the current y
		// as the max of our search; otherwise, we can use it as the min of our search.
		if(bestAlign > align) {
			writeLayout(map);
			return true;
		}
		
		boolean improved = true;
		while(improved) {
			improved = false;
			for(double jump = maxY - minY; !improved && jump > 2; jump /= 2) {
				for(double y = minY + jump / 2; !improved && y < maxY; y += jump) {
					map = tryLayout(y);
					if(isFullLayout(map)) {
						double actualAlign = (getMapTop(map) - top) / (getMapTop(map) - top + bottom - getMapBottom(map));
						if(Math.abs(actualAlign - align) < Math.abs(bestAlign - align)) {
							bestAlign = actualAlign;
							bestMap = map;
							if(bestAlign > align) {
								maxY = y;
							} else {
								minY = y;
							}
							improved = true;
						}
					}
				}
			}
		}
		writeLayout(bestMap);
		return true;
	}
	
	private class WordInfo {
		public int newlines = 0;
		public int offset;
		public int totalOffset = 0;
		public TextRun run;
		public double width;
		public double height;
		public boolean exists = true;
	}
	
	private WordInfo getFirstWord(TextRun run, int offset) {
		WordInfo info = new WordInfo();
		
		String text = "";
		while(run != null) {
			text = run.getText();
			while(offset < text.length() && Character.isWhitespace(text.charAt(offset))) {
				if(text.charAt(offset) == '\n') info.newlines++;
				offset++;
			}
			
			info.run = run;
			info.offset = offset;
			info.totalOffset += offset;
			
			if(offset == text.length()) {
				run = run.getNextRun();
				offset = 0;
			} else {
				break;
			}
		}
		if(run == null) {
			info.exists = false;
			return info;
		}
		
		int i = offset;
		while(i < text.length() && !Character.isWhitespace(text.charAt(i))) i++;
		
		// first word
		text = text.substring(offset, i);
		
		AttributedString astr = new AttributedString(text);
		astr.addAttribute(TextAttribute.FONT, run.getFont());
		
		TextLayout layout = new TextLayout(astr.getIterator(), context);
		
		info.width = layout.getBounds().getWidth() + 2 * endPadding + .1;
		info.height = layout.getAscent() + layout.getDescent();
		
		return info;
	}
	
	// Expand the left and right edges of the given rectangle as much as possible while still keeping
	// it within the shape. It must be contained within the shape as it was given in order for this
	// to work. Do not go beyond the minimum and maximum x values given.
	// O(log(width))
	private void expand(Rectangle2D rect, double minX, double maxX) {
		double y = rect.getMinY(), h = rect.getHeight();
		
		// Use a binary search on one side, then the other
		double x = rect.getMinX(), endX = rect.getMaxX();
		
		while(x - minX > MIN_ERROR) {
			double mid = (x + minX) / 2;
			rect.setRect(mid, y, endX - mid, h);
			if(shape.contains(rect)) {
				x = mid;
			} else {
				minX = mid;
			}
		}
		
		// Now we will keep x as the left end (since it is guarenteed to be in the shape),
		// but search for an x for the right end.
		endX = x;
		x = rect.getMaxX();
		
		while(maxX - x > MIN_ERROR) {
			double mid = (x + maxX) / 2;
			rect.setRect(endX, y, mid - endX, h);
			if(shape.contains(rect)) {
				x = mid;
			} else {
				maxX = mid;
			}
		}
		rect.setRect(endX, y, x - endX, h);
	}
	
	// Given a rectangle including an entire section of a line, return a list of Rectagle2D objects within
	// that line that are inside of the shape. The rectangles will be listed from left to right, and they
	// will all have the same top and height as line, and will be inside both it and the shape.
	// O(log^2 (width)), worst-case O(width)
	private ArrayList getLine(Rectangle2D line) {
		// Do a breadth-first binary search until you find a location in this line where a rectangle will fit
		// with the height of the given line and a width of 1.
		Rectangle2D rect = new Rectangle2D.Double();
		
		// Assume that no text can fit in a space less than 1/5 of the text size
		double minWidth = line.getHeight() * .2 + 2 * endPadding;
		
		for(double jump = line.getWidth(); jump > minWidth; jump /= 2) {
			for(double x = jump / 2; x <= line.getWidth(); x += jump) {
				rect.setRect(x + line.getMinX(), line.getMinY(), minWidth, line.getHeight());
				if(shape.contains(rect)) {
					expand(rect, line.getMinX(), line.getMaxX());
					
					ArrayList list = getLine(new Rectangle2D.Double(line.getMinX(), line.getMinY(), 
							rect.getMinX() - line.getMinX(), line.getHeight()));
					list.add(rect);
					list.addAll(getLine(new Rectangle2D.Double(rect.getMaxX(), line.getMinY(), 
							line.getMaxX() - rect.getMaxX(), line.getHeight())));
					
					return list;
				}
			}
		}
		
		return new ArrayList();
	}
	
	// Use getLine() to determine whether there is space in the given line to hold a word of the
	// given width.
	// O(log^2 (width)), worst-case O(width)
	private boolean lineHasSpace(Rectangle2D line, double width) {
		ArrayList rects = getLine(line);
		
		for(int i = 0; i < rects.size(); i++) {
			Rectangle2D rect = (Rectangle2D)rects.get(i);
			if(rect.getWidth() >= width) {
				return true;
			}
		}
		return false;
	}
	
	// Find a y coordinate of a line in the shape that can hold a word with the given width and height.
	// Locate a y that is as close as possible to the alignment value requested. An align of -1 would
	// return the possible location closest to the top, 1 would return the possible location closest to
	// the bottom, and 0 would return the possible location closest to the center.
	// Return the whole line that should be first, as a Rectangle, or null of there is no way to place the word.
	// O(log(height) log^2 (width)), worst-case O(height width)
	private Rectangle2D placeWord(double width, double height, double align) {
		// Perform a binary search for the best location
		double minY = shape.getBounds2D().getMinY() + endPadding;
		double maxY = shape.getBounds2D().getMaxY() - height - endPadding;
		align = (1 + align) / 2;
		double bestY = minY * (1 - align) + maxY * (align);
		double left = shape.getBounds2D().getMinX(), lineWidth = shape.getBounds2D().getWidth();
		
		// Find at least one y that will work, using breadth-first search. Then, search for another y that
		// no further than that from the best y. If you do not find *any* y that works, word placement has
		// failed and null is returned; as long as I find *some* better y in each rep, I keep searching.
		Rectangle2D bestLine = null;
		Rectangle2D line = new Rectangle2D.Double();
		
		boolean improved = true;
		while(improved) {
			improved = false;
			for(double jump = maxY - minY; !improved && jump > MIN_ERROR; jump /= 2) {
				for(double y = minY + jump / 2; !improved && y < maxY; y += jump) {
					line.setRect(left, y, lineWidth, height);
					if(lineHasSpace(line, width)) {
						bestLine = line;
						line = new Rectangle2D.Double();
						
						double off = Math.abs(y - bestY);
						minY = Math.max(minY, bestY - off);
						maxY = Math.min(maxY, bestY + off);
						
						improved = true;
					}
				}
			}
		}
		
		return bestLine;
	}
	
	private class LineLayout {
		private TextRun firstRun, lastRun;
		private int totalOffset, endOffset;
		private ArrayList shapes;
		private int status;
		private double growHeight, bottom;
		
		public LineLayout(TextRun firstRun, int startOffset, double height, double top) {
			this.firstRun = firstRun;
			shapes = new ArrayList();
			bottom = top + height;
			doLayout(startOffset, height, top);
		}
		
		public void doLayout(int offset, double height, double top) {
			// Construct one long AttributedString consisting of whole text to be written, past
			// the startOffset, up to any newline. Remember if you ended at a newline rather than 
			// the actual end.
			//
			// Make a LineBreakMeasurer for that string. As you layout each section of the line,
			// you may be shaping parts of multiple runs. So, for each section:
			// 1) If the measurer doesn't advance, skip to the next section.
			// 2) Check the height of the layout returned, and if it is bigger than the current 
			//    height, back up the measurer to where it started this section, and try laying 
			//    out to the various boundaries until you find one where the height doesn't change.
			//    Remember the growHeight of the first section beyond that. Set status to GROW
			// 3) Find a transform that will position the returned layoout with the correct alignment
			//    in this section.
			// 4) Go through each run involved and independently create its shape with a TextLayout,
			//    writing that shape to the shapes list using the transform you just made and creating
			//    the next path when needed.
			// 5) If the measurer is at the end, status is either LINE_END or TEXT_END depending on
			//    whether you ended the text at a a newline or included it all.
			// 6) Continue only as long as the status is WORKING
			
			TextRun run = firstRun;
			
			String text = run.getText().substring(offset);
			int limit;
			while((limit = text.indexOf("\n")) == -1 && (run = run.getNextRun()) != null) {
				text += run.getText();
			}
			if(limit < 0) limit = text.length();
			
			if(limit == 0) {
				status = TEXT_END;
				return;
			}
			
			AttributedString string = new AttributedString(text);
			run = firstRun;
			int endOfRun = run.getText().length() - offset;
			int pos = endOfRun;
			string.addAttribute(TextAttribute.FONT, run.getFont(), 0, endOfRun);
			while((run = run.getNextRun()) != null) {
				int len = run.getText().length();
				if(len > 0) {
					string.addAttribute(TextAttribute.FONT, run.getFont(), pos, pos + len);
					pos += len;
				}
			}
			LineBreakMeasurer measurer = new LineBreakMeasurer(string.getIterator(), context);
			// We will run measurer up to limit. We know the limit is a newline if it's less than
			// the length of the text.
			
			GeneralPath path = new GeneralPath();
			shapes.add(path);
			double align = (1 + alignH) / 2;
			run = firstRun;
			lastRun = firstRun;
			status = WORKING;
			double leading = 0;
			ArrayList rects = getLine(new Rectangle2D.Double(shape.getBounds2D().getMinX(), 
					top, shape.getBounds2D().getWidth(), height));
			for(int i = 0; i < rects.size() && status == WORKING; i++) {
				Rectangle2D rect = (Rectangle2D)rects.get(i);
				float layoutWidth = (float)(rect.getWidth() - 2 * endPadding);
				pos = measurer.getPosition();
				if(pos >= limit || layoutWidth < .2 * height) continue;
				
				TextLayout layout = measurer.nextLayout(layoutWidth, limit, true);
				if(layout == null || pos == measurer.getPosition()) continue;
				
				if(layout.getAscent() + layout.getDescent() > height) {
					// line needs to grow
					status = GROW;
					growHeight = layout.getAscent() + layout.getDescent();
					// Find out where the growth occurred - binary search between pos and position now
					int minPos = pos, maxPos = measurer.getPosition();
					while(maxPos - minPos > 1) {
						int posLimit = (minPos + maxPos) / 2;
						measurer.setPosition(pos);
						layout = measurer.nextLayout(layoutWidth, posLimit, true);
						if(layout.getAscent() + layout.getDescent() > height) {
							growHeight = layout.getAscent() + layout.getDescent();
							maxPos = posLimit;
						} else {
							minPos = posLimit;
						}
					}
					measurer.setPosition(pos);
					layout = measurer.nextLayout(layoutWidth, minPos, true);
					// layout now contains only what we can shape without growing the line
					// growHeight now contains the height of the *first* run needing growing
				}
				double width = layout.getBounds().getWidth();
				leading = Math.max(leading, layout.getLeading());
				
				double padding;
				if(rect.getWidth() - width < 2 * endPadding) {
					padding = (rect.getWidth() - width) / 2;
				} else {
					padding = endPadding;
				}
				
				double x = -layout.getBounds().getMinX() + (1 - align) * (rect.getMinX() + padding) + (align) * (rect.getMaxX() - width - padding);
				double y = rect.getMaxY() - layout.getDescent();
				
				AffineTransform transform = AffineTransform.getTranslateInstance(x, y);
				// This transform is now appropriate for the line. We will add the advance to x as we layout
				// each piece of a run that makes it up.
				totalOffset = measurer.getPosition();
				if(totalOffset > endOfRun) {
					measurer.setPosition(pos);
					layout = measurer.nextLayout(layoutWidth, endOfRun, true);
					while(measurer.getPosition() < totalOffset) {
						// For every run that ends before the current measurer position, make a layout of
						// just that part, store it to the current path, then move on to the next run and
						// make a path for it.
						path.append(layout.getOutline(transform), false);
						
						run = run.getNextRun();
						lastRun = run;
						int len = run.getText().length();
						endOfRun += len;
						
						path = new GeneralPath();
						shapes.add(path);
						transform.translate(layout.getAdvance(), 0);
						
						if(len > 0) {
							layout = measurer.nextLayout(layoutWidth, Math.min(endOfRun, totalOffset), true);
						} else {
							layout = new TextLayout(" ", run.getFont(), context);
						}
					}
				}
				path.append(layout.getOutline(transform), false);
				
				// If I reached the limit of the measurer, figure out what sort of limit it is.
				if(totalOffset >= limit) {
					if(limit == text.length()) {
						status = TEXT_END;
					} else {
						// We have reached a "\n" at the end of a line. Print the appropriate number of
						// newlines based on where the next word after that is. 
						status = LINE_END;
						endOffset = totalOffset + lastRun.getText().length() - endOfRun;
						
						WordInfo next = getFirstWord(lastRun, endOffset);
						
						bottom += leading + (next.newlines - 1) * (height / 2 + leading);
						lastRun = next.run;
						endOffset = next.offset;
						totalOffset += next.totalOffset;
						
						return;
					}
				}
			}
			if(status == WORKING) status = LINE_END;
			
			bottom += leading;
			endOffset = totalOffset + lastRun.getText().length() - endOfRun;
			
			//System.out.println("End offset: " + endOffset);
		}
		
		public void writeToMap(HashMap map) {
			TextRun run = firstRun;
			for(int i = 0; i < shapes.size(); i++, run = run.getNextRun()) {
				GeneralPath path = (GeneralPath)map.get(run);
				if(path == null) {
					path = new GeneralPath();
					map.put(run, path);
				}
				path.append((GeneralPath)shapes.get(i), false);
			}
		}
		
		public int getStatus() {
			return status;
		}
		
		public double getGrowHeight() {
			return growHeight;
		}
		
		public int getTotalOffset() {
			return totalOffset;
		}
		
		public int getEndOffset() {
			return endOffset;
		}
		
		public TextRun getLastRun() {
			return lastRun;
		}
		
		public double getBottom() {
			return bottom;
		}
	}
	
	private LineLayout layoutLine(TextRun run, int offset, double top) {
		WordInfo first = getFirstWord(run, offset);
		run = first.run;
		offset = first.offset;
		
		LineLayout oldLine = new LineLayout(run, offset, first.height, top);
		while(oldLine.getStatus() == GROW) {
			LineLayout newLine = new LineLayout(run, offset, oldLine.getGrowHeight(), top);
			if(newLine.getTotalOffset() < oldLine.getTotalOffset()) {
				return oldLine;
			} else {
				oldLine = newLine;
			}
		}
		return oldLine;
	}
	
	private HashMap tryLayout(double minY) {
		double maxY = shape.getBounds2D().getMaxY() - endPadding;
		HashMap map = new HashMap();
		map.put("Top", new Double(minY));
		int offset = 0, totalChars = 0;
		TextRun run = start;
		double bottom = minY;
		
		while(minY < maxY) {
			LineLayout line = layoutLine(run, offset, minY);
			if(line.getLastRun() != run || line.getEndOffset() != offset) bottom = line.getBottom();
			line.writeToMap(map);
			
			if(line.getStatus() == TEXT_END) {
				map.put("Bottom", new Double(bottom));
				return map;
			}
			
			offset = line.getEndOffset();
			minY = line.getBottom();
			run = line.getLastRun();
			totalChars += line.getTotalOffset();
		}
		
		map.put("Bottom", new Double(bottom));
		map.put("Count", new Integer(totalChars));
		return map;
	}
	
	private double getMapBottom(HashMap map) {
		Double b = (Double)map.get("Bottom");
		if(b == null) return Double.NaN;
		
		return b.doubleValue();
	}
	
	private double getMapTop(HashMap map) {
		return ((Double)map.get("Top")).doubleValue();
	}
	
	private int getMapLength(HashMap map) {
		return ((Integer)map.get("Count")).intValue();
	}
	
	private boolean isFullLayout(HashMap map) {
		return map.get("Count") == null;
	}
	
	private void writeLayout(HashMap map) {
		writeLayout(map, null);
	}
	
	private void writeLayout(HashMap map, AffineTransform transform) {
		for(TextRun run = start; run != null; run = run.getNextRun()) {
			Shape path;
			
			if(map == null || (path = (GeneralPath)map.get(run)) == null) path = new GeneralPath();
			else if(transform != null) path = transform.createTransformedShape(path);
			
			run.setShape(path);
		}
	}
}
