package objectdraw;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.BasicStroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.GeneralPath;
import java.awt.geom.NoninvertibleTransformException;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * A <code>Turtle</code> object allows you to draw lines and filled-in polygons by giving
 * instructions to a computer creature called a "turtle" that moves around on the screen and
 * draws a line as it goes. The instructions are of the form "move forward this far" or "turn
 * right this many degrees". Often it is easier to draw graphics this way than it is to try to
 * figure out the coordinates of all the points involved. So, for example, I can draw a pentagon
 * without needing any trigonometry, because all it takes to walk out a pentagon is to move
 * forward, turn right 72&deg;, and repeat that five times. If I want to fill in that pentagon,
 * all I need to do is to tell the turtle to <code>beginFill()</code> when it starts drawing,
 * and then to <code>close()</code> the shape and <code>endFill()</code> when it is done.
 * <br><br>
 * Once a turtle has completed its sketch, you can move that sketch around and resize it just as
 * you would any object on the canvas. You can also rotate a turtle sketch. Resizing and
 * rotating are centered around the point at which the turtle was located when it was first
 * created.
 * <br><br>
 * Because the sketch may contain many different colors - I can change the pen color as I move, 
 * and each fill area has its own color - the <code>setColor()</code> method would not work to
 * recolor the sketch after it is done. Instead, there is a method <code>recolorSketch()</code>
 * that can be used to take all parts of the sketch that were originally in a particular color,
 * and change them to another color.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 *
 */
public class Turtle extends ObjectDrawShape implements DrawableStrokeInterface, LocatableInterface {
	private BasicStroke stroke;
	private boolean pen;
	private TurtleFill fill;
	private TurtleLine line;
	private ArrayList fills, lines;
	private double heading;
	private boolean turtleHidden;
	private Location internalLocation;
	private AffineTransform sketchTransform;
	private double sketchScale;
	private double sketchRotation;
	private Location sketchOrigin;
	private HashMap recolorings;
	
	private static final int STEP_SIZE = 40;
	private static final Shape DEFAULT_SHAPE = defaultShape();
	private static double[] t = new double[4];

	private abstract class TurtleDrawable {
		public abstract void addShape(Area shape);
		public abstract void draw(Graphics2D g);
	}

	// Create a new line only when the turtle changes its Color or Stroke.
	private class TurtleLine extends TurtleDrawable {
		private Color color;
		private GeneralPath path;
		private BasicStroke stroke;
		
		public TurtleLine(TurtleLine clone) {
			color = clone.color;
			stroke = clone.stroke;
			lines.add(this);
			path = new GeneralPath(clone.path);
		}

		public TurtleLine() {
			this.color = getColor();
			this.stroke = Turtle.this.stroke;
			lines.add(this);
			path = new GeneralPath();
			path.moveTo((float)internalLocation.getDoubleX(), (float)internalLocation.getDoubleY());
		}
		
		public void move() {
			if(pen) {
				path.lineTo((float)internalLocation.getDoubleX(), (float)internalLocation.getDoubleY());
			} else {
				path.moveTo((float)internalLocation.getDoubleX(), (float)internalLocation.getDoubleY());
			}
		}

		public void draw(Graphics2D g) {
			translateColor(g, color);
			g.setStroke(stroke);
			g.draw(path);
		}

		public void addShape(Area shape) {
			shape.add(new Area(stroke.createStrokedShape(path)));
		}
		
		public void close() {
			path.closePath();
		}
	}

	private class TurtleFill extends TurtleDrawable {
		private GeneralPath path;
		private TurtleFill next;
		private boolean outline;
		private Color fillColor;
		
		public TurtleFill(TurtleFill clone) {
			path = new GeneralPath(clone.path);
			fillColor = clone.fillColor;
			
			if(clone.next != null) {
				next = new TurtleFill(clone.next);
			}
			fills.add(this);
		}

		public TurtleFill(Color fillColor) {
			path = new GeneralPath();
			path.moveTo((float)internalLocation.getDoubleX(), (float)internalLocation.getDoubleY());

			next = fill;
			fill = this;

			this.fillColor = fillColor;
			
			fills.add(this);
		}

		public void pop() {
			fill =  next;
		}

		public void moved() {
			if(pen) {
				path.lineTo((float)internalLocation.getDoubleX(), (float)internalLocation.getDoubleY());
			} else {
				path.closePath();
				path.moveTo((float)internalLocation.getDoubleX(), (float)internalLocation.getDoubleY());
			}
		}

		public void draw(Graphics2D g) {
			translateColor(g, fillColor);
			g.fill(path);
		}

		public boolean contains(Location point) {
			return path.contains(point.getDoubleX(), point.getDoubleY());
		}

		public void addShape(Area shape) {
			shape.add(new Area(path));
		}
	}
	
	private class TurtleLocation extends Location {
		public TurtleLocation(double x, double y) {
			super(x, y);
		}
		
		public void translate(double dx, double dy) {
			// Called when the Turtle is moved by LocatableInterface methods
			super.translate(dx, dy);
			
			if(this == sketchOrigin) {
				rebuildTransform();
			} else {
				// Send this translation backwards through sketchTransform to figure out
				// what it will do internally
				synchronized(t) {
					t[0] = dx;
					t[1] = dy;
					try {
						sketchTransform.createInverse().transform(t, 0, t, 2, 1);
						moveInternal(t[2] + sketchTransform.getTranslateX(), t[3] + sketchTransform.getTranslateY());
					} catch (NoninvertibleTransformException e) {
					}
				}
			}
		}
	}
	
	/**
	 * A little triangle-shaped icon represents the turtle while it is drawing the sketch, to
	 * make it easier for you to see where the turtle is and which way it is facing when you are
	 * trying to figure out how to get it to draw a shape. You can hide that icon and leave just
	 * the sketch visible by using this method.
	 */
	public void hideTurtle() {
		turtleHidden = true;
		update();
	}
	
	/**
	 * Show the turtle icon again after it had been hidden
	 */
	public void showTurtle() {
		turtleHidden = false;
		update();
	}
	
	/**
	 * Determine whether the turtle's icon is hidden
	 * @return whether the turtle's icon is hidden
	 */
	public boolean isTurtleHidden() {
		return turtleHidden;
	}
	
	/**
	 * Close the shape that the turtle is currently drawing by connecting a line back to the
	 * starting point. The starting point may not be the original starting point of the turtle;
	 * a new "path" is started when you change the line color or stroke.
	 */
	public void close() {
		line.close();
		line = null;
	}
	
	/**
	 * Tell the turtle to lift up its pen, so that it moves without drawing a line behind it.
	 * It can still outline fill areas with its movements.
	 */
	public void penUp() {
		pen = false;
		line = null;
		update();
	}
	
	/**
	 * Tell the turtle to resume drawing a line behind it as it moves.
	 */
	public void penDown() {
		pen = true;
		update();
	}
	
	/**
	 * Determine whether the turtle is leaving a line behind as it moves
	 * @return whether the turtle is leaving a line behind as it moves
	 */
	public boolean isPenDown() {
		return pen;
	}
	
	/**
	 * Get the turtle's heading - the direction, as a degree angle, that it is facing. 
	 * @return the turtle's heading
	 */
	public double getHeading() {
		return heading + sketchRotation;
	}
	
	/**
	 * Give the turtle a new heading
	 * @param heading the degree angle toward which you want the turtle to face
	 */
	public void setHeading(double heading) {
		this.heading = (heading - sketchRotation) % 360.;
		update();
	}
	
	/**
	 * Turn 90&deg; to the right.
	 */
	public void rt() {
		rt(90);
	}
	
	/**
	 * Turn <code>angle</code> degrees to the right
	 * @param angle number of degrees to turn
	 */
	public void rt(double angle) {
		setHeading(getHeading() - angle);
	}
	
	/**
	 * Turn 90&deg; to the left.
	 */
	public void lt() {
		lt(90);
	}
	
	/**
	 * Turn <code>angle</code> degrees to the left
	 * @param angle number of degrees to turn
	 */
	public void lt(double angle) {
		setHeading(getHeading() + angle);
	}
	
	/**
	 * Move forward 40 pixels
	 */
	public void move() {
		move(STEP_SIZE);
	}
	
	/**
	 * Move forward <code>distance</code> pixels
	 * @param distance number of pixels to move
	 */
	public void move(double distance) {
	   	double ang = -heading * Math.PI / 180.;
    	
    	moveInternal(distance * Math.cos(ang), distance * Math.sin(ang));
 	}

	/**
	 * Construct a <code>Turtle</code> at (<code>x</code>, <code>y</code>), facing to the right,
	 * and drawing a line in the given <code>color</code>.
	 * @param x x coordinate of the turtle
	 * @param y y coordinate of the turtle
	 * @param color <code>Color</code> the turtle will be drawing in
	 * @param canvas <code>DrawingCanvas</code> on which the turtle should be placed
	 */
	public Turtle(double x, double y, Color color, DrawingCanvas canvas) {
		super(null, false, color, canvas);
		
		internalLocation = new Location(0, 0);
		setMyLocation(new TurtleLocation(x, y));
		pen = true;
		recolorings = new HashMap();
		heading = 0;
		stroke = DEFAULT_STROKE;
		turtleHidden = false;
		
		sketchOrigin = new TurtleLocation(x, y);
		sketchScale = 1;
		sketchRotation = 0;
		sketchTransform = new AffineTransform();
		
		clearSketch();
		rebuildTransform();
		
		ready();
	}

	/**
	 * Construct a <code>Turtle</code> at <code>start</code>, facing to the right,
	 * and drawing a line in the given <code>color</code>.
	 * @param start the <code>Location</code> at which the turtle should be placed
	 * @param color <code>Color</code> the turtle will be drawing in
	 * @param canvas <code>DrawingCanvas</code> on which the turtle should be placed
	 */
	public Turtle(Location start, Color color, DrawingCanvas canvas) {
		this(start.getDoubleX(), start.getDoubleY(), color, canvas);
	}

	/**
	 * Construct a <code>Turtle</code> at (<code>x</code>, <code>y</code>), facing to the right,
	 * and drawing a line in black.
	 * @param x x coordinate of the turtle
	 * @param y y coordinate of the turtle
	 * @param canvas <code>DrawingCanvas</code> on which the turtle should be placed
	 */
	public Turtle(double x, double y, DrawingCanvas canvas) {
		this(x, y, null, canvas);
	}

	/**
	 * Construct a <code>Turtle</code> at <code>start</code>, facing to the right,
	 * and drawing a line in the given <code>color</code>.
	 * @param start the <code>Location</code> at which the turtle should be placed
	 * @param canvas <code>DrawingCanvas</code> on which the turtle should be placed
	 */
	public Turtle(Location start, DrawingCanvas canvas) {
		this(start, null, canvas);
	}

	/**
	 * Construct a <code>Turtle</code> at the center of <code>canvas</code>, facing to the right,
	 * and drawing a line in the given <code>color</code>.
	 * @param color <code>Color</code> the turtle will be drawing in
	 * @param canvas <code>DrawingCanvas</code> on which the turtle should be placed
	 */
	public Turtle(Color color, DrawingCanvas canvas) {
		this(canvas.getWidth() / 2, canvas.getHeight() / 2, color, canvas);
	}

	/**
	 * Construct a <code>Turtle</code> at the center of <code>canvas</code>, facing to the right,
	 * and drawing a line in black.
	 * @param canvas <code>DrawingCanvas</code> on which the turtle should be placed
	 */
	public Turtle(DrawingCanvas canvas) {
		this((Color)null, canvas);
	}

	/**
	 * Construct a clone of <code>clone</code> and place it on the given canvas. This allows you
	 * to sketch a shape once with a single <code>Turtle</code> and then use that as a template
	 * to make more copies of that shape.
	 * @param clone <code>Turtle</code> to copy
	 * @param canvas <code>DrawingCanvas</code> on which the new turtle should be placed
	 */
	public Turtle(Turtle clone, DrawingCanvas canvas) {
		super(null, false, clone.getColor(), canvas);
		
		internalLocation = clone.internalLocation;
		setMyLocation(new TurtleLocation(clone.getDoubleX(), clone.getDoubleY()));
		pen = clone.pen;
		
		ArrayList activeFills = new ArrayList();
		fill = new TurtleFill(clone.fill);
		fills = new ArrayList();
		for(int i = 0; i < clone.fills.size() - activeFills.size(); i++) {
			new TurtleFill((TurtleFill)clone.fills.get(i));
		}
		fills.addAll(activeFills);
		
		lines = new ArrayList();
		for(int i = 0; i < clone.lines.size(); i++) {
			new TurtleLine((TurtleLine)clone.lines.get(i));
		}
		if(clone.line != null) line = (TurtleLine)lines.get(lines.size() - 1);
		
		recolorings = new HashMap(clone.recolorings);
		heading = clone.heading;
		stroke = clone.stroke;
		turtleHidden = clone.turtleHidden;
		
		sketchOrigin = new TurtleLocation(clone.getSketchDoubleX(), clone.getSketchDoubleY());
		sketchScale = clone.sketchScale;
		sketchRotation = clone.sketchRotation;
		sketchTransform = clone.sketchTransform;
		update();
	}

	private void moveInternal(double dx, double dy) {
		// Called to move Turtle within its own coordinate system
		if(line == null) line = new TurtleLine();
		internalLocation.translate(dx, dy);
		line.move();
		
		if(fill != null) fill.moved();
		
		update();
	}
	
	private void translateColor(Graphics2D g, Color c) {
		Color translation = (Color)recolorings.get(c);
		if(translation == null) {
			g.setColor(c);
		} else {
			g.setColor(translation);
		}
	}
	
	/**
	 * @deprecated for internal use only
	 */
	public void draw(Graphics2D g) {
		AffineTransform saveAT = g.getTransform();
		
		g.transform(sketchTransform);
		if(lines == null) return;
		
		for(int i = 0; i < fills.size(); i++) {
			TurtleDrawable draw = (TurtleDrawable)fills.get(i);
			draw.draw(g);
		}
		
		for(int i = 0; i < lines.size(); i++) {
			TurtleDrawable draw = (TurtleDrawable)lines.get(i);
			draw.draw(g);
		}
		
		if(!turtleHidden) {
			Shape shape = makeTurtleShape();
			g.setStroke(stroke);
			translateColor(g, getColor());
			
			if(pen) g.fill(shape);
			
			g.draw(shape);
		}
		g.setTransform(saveAT);
	}
	
	private Shape makeTurtleShape() {
	   	double ang = heading * Math.PI / 180.;
    	double cos = Math.cos(ang), sin = Math.sin(ang);
    	
    	AffineTransform n = new AffineTransform(cos, -sin, sin, cos, 
    			internalLocation.getDoubleX(), internalLocation.getDoubleY());
 		return n.createTransformedShape(DEFAULT_SHAPE);
	}

	/**
	 * @deprecated for internal use only
	 */
	public static Shape defaultShape() {
		GeneralPath me = new GeneralPath();

		me.moveTo(0, 0);
		me.lineTo(-4, -8);
		me.lineTo(12, 0);
		me.lineTo(-4, 8);
		me.closePath();

		return me;
	}
	
	/**
	 * Tell this turtle to begin outlining a shape that will be filled in with 
	 * <code>fillColor</code>. You should later call <code>endFill()</code> once the shape
	 * has been outlined. If a new fill region is started with another call to 
	 * <code>beginFill()</code> before you call <code>endFill()</code>, the turtle will start
	 * adding to the new fill region instead and will make no changes to the previous fill
	 * region until the new one is ended.
	 * @param fillColor <code>Color</code> to fill in with
	 */
	public void beginFill(Color fillColor) {
		new TurtleFill(fillColor);
	}
	
	/**
	 * Finish up the fill region you are currently working on, and go back to adding to the
	 * previously active fill region, if there is one.
	 */
	public void endFill() {
		if(fill != null) fill.pop();
	}

	/**
	 * @deprecated for internal use only
	 */
	public Shape makeShape() {
		Area sketchShape = new Area();
		
		for(int i = 0; i < fills.size(); i++) {
			TurtleDrawable draw = (TurtleDrawable)fills.get(i);
			
			draw.addShape(sketchShape);
		}
		
		for(int i = 0; i < lines.size(); i++) {
			TurtleDrawable draw = (TurtleDrawable)lines.get(i);
			
			draw.addShape(sketchShape);
		}
		return sketchTransform.createTransformedShape(sketchShape);
	}
	
	/**
	 * Clear this turtle's sketch. All its lines and fills will be deleted. The turtle remains
	 * at its current position.
	 */
	public void clearSketch() {
		fills = new ArrayList();
		lines = new ArrayList();
		fill = null;
		line = null;
		update();
	}
	
	/**
	 * Set the angle, in degrees, by which this sketch should be rotated. Rotation occurs
	 * centered around the starting point of the turtle.
	 * @param angle angle of rotation of the sketch
	 */
	public void setSketchRotation(double angle) {
		sketchRotation = angle % 360.;
		rebuildTransform();
	}
	
	/**
	 * Rotate this sketch by the given angle, in degrees. This adds on to any previous rotation.
	 * @param byAngle angle to rotate by. Positive rotates left, negative rotates right.
	 */
	public void rotateSketch(double byAngle) {
		setSketchRotation(sketchRotation + byAngle);
	}
	
	/**
	 * Get the angle of rotation of this turtle's sketch
	 * @return the angle of rotation
	 */
	public double getSketchRotation() {
		return sketchRotation;
	}
	
	/**
	 * Move the turtle <b>and</b> its entire sketch by the given amount in the x and y.
	 * @param dx distance to shift the sketch in the x direction
	 * @param dy distance to shift the sketch in the y direction
	 */
	public void moveSketch(double dx, double dy) {
		sketchOrigin.translate(dx, dy);
	}
	
	/**
	 * Move the origin on the sketch to (<code>x</code>, <code>y</code>). The origin is the point
	 * on the sketch where the turtle started out.
	 * @param x x coordinate to move the origin to
	 * @param y y coordinate to move the origin to
	 */
	public void moveSketchTo(double x, double y) {
		moveSketch(x - sketchOrigin.getDoubleX(), y - sketchOrigin.getDoubleY());
	}
	
	/**
	 * Move the origin on the sketch to the given <code>Location</code>. The origin is the point 
	 * on the sketch where the turtle started out.
	 * @param point <code>Location</code> to move the origin to
	 */
	public void moveSketchTo(Location point) {
		moveSketchTo(point.getDoubleX(), point.getDoubleY());
	}
	
	/**
	 * Get the x coordinate of the origin of the sketch, in <code>int</code> precision. The 
	 * origin is the point on the sketch where the turtle started out.
	 * @return the x coordinate of the origin of the sketch, in <code>int</code> precision
	 */
	public int getSketchX() {
		return (int)(sketchOrigin.getDoubleX() + .5);
	}
	
	/**
	 * Get the y coordinate of the origin of the sketch, in <code>int</code> precision. The 
	 * origin is the point on the sketch where the turtle started out.
	 * @return the y coordinate of the origin of the sketch, in <code>int</code> precision
	 */
	public int getSketchY() {
		return (int)(sketchOrigin.getDoubleY() + .5);
	}
	
	/**
	 * Get the x coordinate of the origin of the sketch, in <code>double</code> precision. The 
	 * origin is the point on the sketch where the turtle started out.
	 * @return the x coordinate of the origin of the sketch, in <code>double</code> precision
	 */
	public double getSketchDoubleX() {
		return sketchOrigin.getDoubleX();
	}
	
	/**
	 * Get the y coordinate of the origin of the sketch, in <code>double</code> precision. The 
	 * origin is the point on the sketch where the turtle started out.
	 * @return the y coordinate of the origin of the sketch, in <code>double</code> precision
	 */
	public double getSketchDoubleY() {
		return sketchOrigin.getDoubleY();
	}
	
	/**
	 * Get the location of the origin of the sketch. The 
	 * origin is the point on the sketch where the turtle started out.
	 * @return the origin of the sketch in a <code>Location</code> object
	 */
	public Location getSketchLocation() {
		return sketchOrigin;
	}
	
	/**
	 * Set the scale factor of this turtle's sketch, causing it to be shrunk or expanded.
	 * A scale factor of 1 dispalys the sketch at true size, a factor of .5 displays it at
	 * half size, a factor of 3 makes it three times bigger, and so on. The scaling is 
	 * centered around the starting point of the turtle.
	 * 
	 * @param factor factor to which the sketch should be scaled
	 */
	public void setSketchScale(double factor) {
		sketchScale = factor;
		rebuildTransform();
	}
	
	/**
	 * Add a scale factor to this turtle's sketch, causing it to be shrunk or expanded.
	 * A scale factor of .5 shrinks it to half size, a factor of 3 makes it three times bigger, 
	 * and so on. The scaling is centered around the starting point of the turtle. This effect
	 * adds on to whatever scaling you have already done.
	 * 
	 * @param byFactor factor by which the sketch should be scaled
	 */
	public void scaleSketch(double byFactor) {
		setSketchScale(sketchScale * byFactor);
	}
	
	/**
	 * Get the scale factor of the sketch - the degree to which it has been shrunk or expanded.
	 * A factor of 2 means it is twice as big as it was drawn; .25 means it is a quarter of the 
	 * size it was drawn at; 1 means it is exactly the size it started, and so on.
	 * @return the scale factor
	 */
	public double getSketchScale() {
		return sketchScale;
	}
	
	/**
	 * Replace all parts of the sketch that were drawn in the color <code>original</code> with
	 * the new color <code>substitute</code>.
	 * @param original <code>Color</code> with which a portion of the sketch was originally drawn
	 * @param substitute <code>Color</code> to replace that portion of the sketch with
	 */
	public void recolorSketch(Color original, Color substitute) {
		recolorings.put(original, substitute);
	}
	
	private void rebuildTransform() {
		sketchTransform.setToIdentity();
		sketchTransform.translate(sketchOrigin.getDoubleX(), sketchOrigin.getDoubleY());
		sketchTransform.rotate(-sketchRotation * Math.PI / 180);
		sketchTransform.scale(sketchScale, sketchScale);
		update();
	}
	
	/* * * * Methods from DrawableStrokeInterface * * * */

	/**
	 * Return the width of this object's lines.
	 * @return the width of this object's lines
	 */
	public double getLineWidth() {
		return stroke.getLineWidth();
	}

	/**
	 * Return the stroke used for this object's lines.
	 * @return the stroke used for this object's lines
	 */
	public BasicStroke getStroke() {
		return stroke;
	}

	/**
	 * Set the width of this object's lines
	 * @param width the new width of this object's lines
	 */
	public void setLineWidth(double width) {
		stroke = new BasicStroke((float)width, stroke.getEndCap(), stroke.getLineJoin(), 
				stroke.getMiterLimit(), stroke.getDashArray(), stroke.getDashPhase());
		line = null;
		update();
	}

	/**
	 * Set the stroke used for this object's lines
	 * @param stroke the new stroke for this object's lines
	 */
	public void setStroke(BasicStroke stroke) {
		this.stroke = stroke;
		line = null;
		update();
	}
	
	/* * * * Methods from LocatableInterface * * * */
	
	/**
	 * Get this object's x coordinate.
	 * @return this object's x coordinate
	 */
	public int getX() {
		return getLocation().getX();
	}
	/**
	 * Get this object's y coordinate.
	 * @return this object's y coordinate
	 */
	public int getY() {
		return getLocation().getY();
	}
	/**
	 * Get this object's x coordinate, in double precision.
	 * @return this object's x coordinate, in double precision
	 */
	public double getDoubleX() {
		return getLocation().getDoubleX();
	}
	/**
	 * Get this object's y coordinate, in double precision.
	 * @return this object's y coordinate, in double precision
	 */
	public double getDoubleY() {
		return getLocation().getDoubleY();
	}
	/**
	 * Get this object's location. Translating this <code>Location</code> object
	 * will have the same effect as moving the object. So, the following two lines
	 * of code are equivalent:
	 * <br><br><code>
	 * object.move(10, 20);<br>
	 * object.getLocation().translate(10, 20);
	 * </code><br><br>If you want to create a <code>Location</code> offset by some
	 * amount from the <code>Location</code> of this object, without actually
	 * moving this object, use the <code>offset()</code> method in <code>Location</code>,
	 * which creates a new <code>Location</code>.
	 * @return the <code>Location</code> of this object.
	 */
	public Location getLocation() {
		return getMyLocation();
	}

	/* * * * Methods from DrawableInterface * * * */
	
	/**
	 * Make this object invisible. It stays in position on its canvas; it is simply not drawn.
	 * This method has no effect if the object was already hidden.
	 */
	public void hide() { super.hide(); }
	/**
	 * Make an object visible again after it had been hidden. This method has no effect if the
	 * object was already visible.
	 */
	public void show() { super.show(); }
	/**
	 * Return <code>true</code> if this object has been rendered invisible with a call to its
	 * <code>hide()</code> method, <code>false</code> if it is still being drawn.
	 * @return whether this object is hidden
	 */
	public boolean isHidden() { return super.isHidden(); }

	/**
	 * Place this object on the specified canvas. The object will be removed from its current
	 * canvas, if it is on a canvas that is not the one it is being moved to. The object will
	 * be at the front of the canvas, as if it had just been drawn. If the given canvas is 
	 * <code>null</code> the object is simply removed from its current canvas.
	 * @param c <code>DrawingCanvas</code> to place this object on
	 */
	public void addToCanvas(DrawingCanvas c) { super.addToCanvas(c); }
	/**
	 * Remove this object from its canvas.
	 */
	public void removeFromCanvas() { super.removeFromCanvas(); }
	/**
	 * Return the canvas that this object is on. If it has been removed from its canvas, this
	 * method will return <code>null</code>.
	 * @return the <code>DrawingCanvas</code> that this object is on.
	 */
	public DrawingCanvas getCanvas() { return super.getCanvas(); }

	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param point <code>Location</code> to move to
	 */
	public void moveTo(Location point) { super.moveTo(point); }
	/**
	 * Move the reference point of this object to the given location. For a Drawable2D object,
	 * the reference point is the top left corner; for a line, it is the start point.
	 * @param x x coordinate to move to
	 * @param y y coordinate to move to
	 */
	public void moveTo(double x, double y) { super.moveTo(x, y); }
	/**
	 * Shift (translate) this object left or right by an amount <code>dx</code>, and up or
	 * down by an amount <code>dy</code>. A negative <code>dx</code> indicates a shift to the
	 * left; a negative <code>dy</code> indicates a shift up
	 * @param dx amount of x translation
	 * @param dy amount of y translation
	 */
	public void move(double dx, double dy) { super.move(dx, dy); }

	/**
	 * Return the color of this object
	 * @return the <code>Color</code> of this object
	 */
	public Color getColor() { return super.getColor(); }
	/**
	 * Change the color of this object.
	 * @param c new <code>Color</code> for this object
	 */
	public void setColor(Color c) {
		line = null;
		super.setColor(c);
	}

	/**
	 * Move this object one step forward in the draw order, causing it to be drawn on top of
	 * the objects that are behind it.
	 */
	public void sendForward() { super.sendForward(); }
	/**
	 * Move this object one step backward in the draw order, causing it to be drawn underneath
	 * the objects that are in front of it.
	 */
	public void sendBackward() { super.sendBackward(); }
	/**
	 * Move this object to the very front of the draw order, causing it to be drawn on top of
	 * all the other objects.
	 */
	public void sendToFront() { super.sendToFront(); }
	/**
	 * Move this object to the very back of the draw order, causing it to be drawn underneath
	 * all the other objects.
	 */
	public void sendToBack() { super.sendToBack(); }

	/**
	 * Return <code>true</code> if the given location is inside this object, <code>false</code>
	 * otherwise. Because <code>Line</code>s and <code>FramedArc</code>s have no "inside", a
	 * point is considered to be contained in one of them as long as it is within a few pixels
	 * of the line.
	 * @param point <code>Location</code> to test
	 * @return whether the given location is inside this shape
	 */
	public boolean contains(Location point) {
		return getShape().contains(point.getDoubleX(), point.getDoubleY());
	}

	/**
	 * Return a string describing how this object might be constructed.
	 */
	public String toString() {
		return toString(Text.formatDecimal(getDoubleX()) + ", " + 
				Text.formatDecimal(getDoubleY()));
	}
}
