package objectdraw;

import java.awt.BorderLayout;

/**
 * <code>WindowController</code> in the class that defines how you interact with the drawing
 * canvas of your program. It has methods that will be called when the program is first run, 
 * and then when various mouse and keyboard events occur, and in those methods you can use the
 * <code>canvas</code> variable, inherited from <code>WindowController</code>, to draw objects
 * on the canvas.
 * 
 * @author Russell Zahniser (russell@zahniser.net)
 */
public class WindowController extends Controller {
	/**
	 * In a class extending <code>WindowController</code>, you can use the <code>canvas</code>
	 * variable to refer to the canvas of that controller.
	 */
	protected DrawingCanvas canvas;
	
	/**
	 * Construct a new WindowController.
	 */
	public WindowController() {
		super();
		
		canvas = new JDrawingCanvas();
		
		new KeyInterpreter(this, canvas);
		new MouseInterpreter(this, canvas);
	}
	
	/**
	 * Retrieve the <code>canvas</code> object of this <code>WindowController</code>.
	 * @return this <code>WindowController</code>'s canvas.
	 */
	public DrawingCanvas getCanvas() {
		return canvas;
	}
	
	/**
	 * @deprecated for internal use only
	 */
	protected void finishLayout() {
		getContentPane().add((JDrawingCanvas)canvas, BorderLayout.CENTER);
		validate();
	    setVisible(true);
		repaint();
	}
	
	/**
	 * @deprecated for internal use only
	 */
	protected void checkIfReady() {
		// This is called before begin() each time the applet starts
		canvas.clear();
		while(canvas.getWidth() <= 0 || canvas.getHeight() <= 0) {
			try {
				Thread.sleep(20);
			} catch (InterruptedException e) {
			}
		}
	}
	
	/**
	 * @deprecated internal use only
	 */
	public void stop() {
		super.stop();
		canvas.clear();
	}
	
	/**
	 * This method is called when the applet first appear on the screen. Use it to set up
	 * anythign that you want to have ready.
	 */
	public void begin() {}
	/**
	 * This method is called when the user clicks on the canvas - that is, presses and releases
	 * the mouse without dragging it in between.
	 * @param point <code>Location</code> of the mouse
	 */
	public void onMouseClick(Location point) {}
	/**
	 * This method is called as the user is dragging the mouse across the canvas - that is,
	 * moving the mouse with the button held down.
	 * @param point <code>Location</code> of the mouse
	 */
	public void onMouseDrag(Location point) {}
	/**
	 * This method is called as the user is moving the mouse across the canvas without the
	 * mouse button held down.
	 * @param point <code>Location</code> of the mouse
	 */
	public void onMouseMove(Location point) {}
	/**
	 * This method is called when the mouse enters the canvas.
	 * @param point <code>Location</code> of the mouse
	 */
	public void onMouseEnter(Location point) {}
	/**
	 * This method is called when the mouse leaves the canvas.
	 * @param point <code>Location</code> of the mouse
	 */
	public void onMouseExit(Location point) {}
	/**
	 * This method is called when the mouse button is pressed.
	 * @param point <code>Location</code> of the mouse
	 */
	public void onMousePress(Location point) {}
	/**
	 * This method is called when the mouse button is released.
	 * @param point <code>Location</code> of the mouse
	 */
	public void onMouseRelease(Location point) {}
}
